#include "defs.h"

struct Connection
{
  int fd;
  int flags;
  char desc[512];
  char hbuffer[BUFSIZE * 2];
};

struct ServiceCheck
{
  char name[BUFSIZE];
  char host[HOSTLEN];
  int port;
  int status;
};

struct Memo
{
  char to[NICKLEN];
  char from[NICKLEN];
  char memo[BUFSIZE];
  time_t time_sent;
  int flags;
  int id;
};

struct HelpTable
{
  char* name;
  char* filename;
  int flag;
};

struct Mem
{
  int dlink_count;
  int nick_count;
  int operators;
} Mem;

struct Scanned
{
  char ip[HOSTLEN];
};

struct NickKill
{
  char nick[NICKLEN];
  time_t since;
};

struct MemoPurge
{
  char to[NICKLEN];
  char from[NICKLEN];
  char memo[BUFSIZE];

  time_t time_sent;
  int flags;
};

/* Connected user structure */
struct User {

    char	    nick[NICKLEN];
    char	    hostname[HOSTLEN];
    char	    realhost[HOSTLEN];
    char	    username[USERLEN];

    char	    gecos[REALLEN];

    unsigned int    flags;
    unsigned int    modes;
    time_t	    ts;

    struct Server   *server;
    struct Nick	    *n;

	dlink_list		channels;
};

/* Connected server structure */
struct Server {

  char	name[HOSTLEN];
  char	desc[BUFSIZE];

};

/* Registered nickname structure */
struct Nick {
    char	    nick[NICKLEN];
    char	    password[MAXPASS];

    unsigned int    flags;
    unsigned int    expire;

    time_t	    last_seen;
    time_t	    registered;

    char	    last_quit[BUFSIZE];
    char	    last_address[HOSTLEN];
    char	    email[BUFSIZE];
    char	    url[BUFSIZE];

    dlink_list	    links;
    dlink_list	    channels;
    dlink_list	    access_hosts;
};

struct LinkedNicks {
  char	master[NICKLEN];
  char	nick[NICKLEN];
};

/* Registered Channel Structure */
struct RChannel {
	char	name[CHANLEN];
	char	topic[TOPICLEN];

	char	pass[MAXPASS];
	char	desc[BUFSIZE];

	int		modes;
	int		flags;

	struct Nick	*founder;
	struct Levels *levels;

	dlink_list	access;
};

struct Levels {
	int		autodeop;
	int		autovoice;
	int		autohop;
	int		autoop;

	int		cmdvoice;
	int		cmdhop;
	int		cmdop;
	int		cmdaccess;
	int		cmdset;
	int		cmdinvite;
	int		cmdunban;
	int		cmdakick;
	int		cmdclear;
	int		superop;
};

struct Access {
	char	nick[NICKLEN+1];
	int		level;
};

/* Channel Access Entry Structure */
struct AccessEntry {
	char	nick[NICKLEN];
	int		level;
};

struct Channel {
	char	name[CHANLEN];

	int		modes;
	long	ts;
	int		usercount;

	dlink_list	users;
};

struct ChanUser {
	char	nick[NICKLEN];
	int		modes;
};


struct settings {
  char	uplink[HOSTLEN];
  int	port;
  char	pass[MAXPASS];

  int	silent;

/* MySQL Settings */
  char	mysql_host[HOSTLEN];
  char	mysql_user[BUFSIZE];
  char	mysql_pass[MAXPASS];
  char	mysql_db[BUFSIZE];
} settings;

struct me {
    /* Local server information */
    char    name[BUFSIZE];
    char    desc[BUFSIZE];
    int	    silent;
    time_t  started;
	time_t	burststart;
    int	    flags;
    int	    hidden;
    int	    privmsg;
    int	    pid;
	int		loglevel;

    /* Uplink information */
    char    uplink[BUFSIZE];
    char    pass[BUFSIZE];
    int	    port;
    int	    reconnect;

    char    adminname[BUFSIZE];
    char    adminemail[BUFSIZE];

    dlink_list	servers;
    dlink_list	operators;

#ifdef USE_MYSQL
    /* MySQL Settings */
    char    mhost[BUFSIZE];
    char    mpass[BUFSIZE];
    char    muser[BUFSIZE];
    char    mdb[BUFSIZE];
    MYSQL   *mptr;
#endif
} me;

/* NickServ Settings */
struct NickServ {
    char    nick[NICKLEN];
    char    user[USERLEN];
    char    host[HOSTLEN];
    char    name[REALLEN];

    long    expire;
    int	    flags;

#ifdef USE_MYSQL
    char    table[BUFSIZE];
#else
    char    dbfile[BUFSIZE];
#endif

} NickServ;

/* ChanServ Settings */
struct ChanServ {
    char    nick[NICKLEN];
    char    user[USERLEN];
    char    host[HOSTLEN];
    char    name[REALLEN];

    long    expire;
    int	    flags;

#ifdef USE_MYSQL
    char    table[BUFSIZE];
	char	access_table[BUFSIZE];
	char	levels_table[BUFSIZE];
#endif
} ChanServ;

/* MemoServ Settings */
struct MemoServ {
    char    nick[NICKLEN];
    char    user[USERLEN];
    char    host[HOSTLEN];
    char    name[REALLEN];

    long    expire;

#ifdef USE_MYSQL
    char    table[BUFSIZE];
#endif
} MemoServ;

struct BotServ {
    char    nick[NICKLEN];
    char    user[USERLEN];
    char    host[HOSTLEN];
    char    name[REALLEN];

    dlink_list	bots;

#ifdef USE_MYSQL
    char    table[BUFSIZE];
	char	channel_table[BUFSIZE];
#endif
} BotServ;

/* OperServ Settings */
struct OperServ {
    char    nick[NICKLEN];
    char    user[USERLEN];
    char    host[HOSTLEN];
    char    name[REALLEN];

    dlink_list	channels;

    int	    operonly;
    long    save;
} OperServ;

struct Operator {
    char    nick[NICKLEN+1];
    char    password[MAXPASS+1];
    int	    flags;

    dlink_list	hosts;
};
