#include "fwserv.h"
#include "config.h"
#include "botserv.h"
#include "user.h"

#if defined(NICKSERVICES) && defined(BOTSERVICES)

void botserv(struct User *, char *, char *);

void AddChannelToBot (struct Bot *, char *);
void RemoveChannelFromBot (struct Bot *, char *);
int BotOnChannel (struct Bot *, char *);

static int bs_help (struct User *, int, char **);
static int bs_add (struct User *, int, char **);
static int bs_del (struct User *, int, char **);
static int bs_change (struct User *, int, char **);
static int bs_join (struct User *, int, char **);
static int bs_part (struct User *, int, char **);
static int bs_list (struct User *, int, char **);
static int bs_unused (struct User *, int, char **);

static struct CmdTable botserv_table[] =
{
  { "HELP",	NULL,	bs_help,	LVL_NONE },

  { "LIST",	NULL,	bs_list,	LVL_NONE },

  { "ADD",	NULL,	bs_add,		LVL_OPER },
  { "DEL",	NULL,	bs_del,		LVL_OPER },
  { "ChANGE",	NULL,	bs_change,	LVL_OPER },

  { "JOIN",	NULL,	bs_join,	LVL_IDENT },
  { "PART",	NULL,	bs_part,	LVL_IDENT },

  { "SET",	NULL,	bs_unused,	LVL_IDENT },

  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0}
};

static struct CmdTable set_table[] =
{
  { "NICK",	NULL,	bs_unused,	LVL_IDENT },
  { "USER",	NULL,	bs_unused,	LVL_IDENT },
  { "HOST",	NULL,	bs_unused,	LVL_IDENT },

  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0 }
};

static struct HelpTable help_table[] =
{
  { "ADD",	"help/botserv/add",	LVL_IDENT },
  { "DEL",	"help/botserv/del",	LVL_IDENT },
  { "CHANGE",	"help/botserv/change",	LVL_IDENT },

  { "JOIN",	"help/botserv/join",	LVL_NONE },
  { "PART",	"help/botserv/part",	LVL_NONE },

  { "SET",	"help/botserv/set",	LVL_IDENT },
};

static struct HelpTable help_set_table[] =
{
  { "NICK",	"help/botserv/set/nick",	LVL_IDENT },
  { "USER",	"help/botserv/set/user",	LVL_IDENT },
  { "HOST",	"help/botserv/set/host",	LVL_IDENT },
};

void botserv(struct User *u, char *target, char *param)
{
    char *parv[MAXPARA];
    int parc, i, status;
    static char buf[BUFSIZE];
    parc = 0;

    for (i = 0; i < MAXPARA; i++)
        parv[i] = NULL;

    strlcpy(buf, param, BUFSIZE);

    if (param != NULL)
        parc = msgtoken(param, parv);

    if (parc == 0)
        return;

    for (i = 0; botserv_table[i].handler; i++)
    {
        if(!strcasecmp(parv[0], botserv_table[i].name))
        {
            if ((botserv_table[i].flag & LVL_IDENT) && !IsIdent(u))
            {
                privmsg(u->nick, BotServ.nick, RPL_MUST_IDENTIFY, botserv_table[i].name);
                return;
            }
            if((botserv_table[i].flag & LVL_ADMIN) && !(u->flags & LVL_ADMIN))
            {
                privmsg(u->nick, BotServ.nick, "Invalid Command: %s", parv[0]);
                return;
            }
            if((botserv_table[i].flag & LVL_OPER) && !(u->flags & LVL_OPER))
            {
                privmsg(u->nick, BotServ.nick,  "Invalid Command: %s", parv[0]);
                return;
            }
            status = botserv_table[i].handler(u, parc, parv);
            return;
        }
    }

    privmsg (u->nick, BotServ.nick, "Invalid Command: %s", parv[0]);
    return;
}

static int
bs_unused (struct User *u, int parc, char *parv[])
{
  return;
}

static int
bs_list (struct User *u, int parc, char *parv[])
{
  struct Bots *bot;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, bot_list.head)
  {
    count++;
    bot = (struct Bots *)dl->data;
    privmsg(u->nick, BotServ.nick, "%d: %s [%s@%s]", count, bot->nick, bot->username, bot->hostname);
  }
}

static int
bs_help (struct User *u, int parc, char *parv[])
{
  int i;

  if (parc < 2)
  {
    send_help_file(u, BotServ.nick, "help/botserv/index", "index");
    return;
  }

  if (!strcasecmp(parv[1], "SET"))
  {
    if (parc < 3)
    {
      send_help_file(u, BotServ.nick, "help/botserv/set/index", "index");
      return;
    }
    for (i=0; help_set_table[i].name; i++)
    {
      if (!strcasecmp(parv[2], help_set_table[i].name))
      {
        send_help_file(u, BotServ.nick, help_set_table[i].filename, parv[2]);
        return;
      }
    }
    return;
  }
  
  for (i=0; help_table[i].name; i++)
  {
    if (!strcasecmp(parv[1], help_table[i].name))
    {
      send_help_file(u, BotServ.nick, help_table[i].filename, parv[1]);
      return;
    }
  }
  privmsg(u->nick, BotServ.nick, "No help for %s", parv[1]);
  return;
}

static int bs_add (struct User *u, int parc, char *parv[])
{
    int ret;

    if (parc < 4)
    {
	privmsg(u->nick, BotServ.nick, "Syntax: \002ADD\002 <nick> <username> <hostname>");
	return;
    }

    ret = AddBot (parv[1], parv[2], parv[3], (parv[4])? parv[4] : "Channel Holder");

    if (ret >= 0)
    {
	privmsg (u->nick, BotServ.nick, "Bot \002%s\002 Created", parv[1]);
	return;
    }

    if (ret == -1)
    {
	privmsg (u->nick, BotServ.nick, "Bot \002%s\002 exists already.", parv[1]);
	return;
    }

    if (ret < -1)
    {
	privmsg (u->nick, BotServ.nick, "Unknown Error Creating Bot");
	return;
    }

    /* NOT REACHED */
    return;
}

static int bs_del (struct User *u, int parc, char *parv[])
{
    struct Bot *bot;

    if (parc < 2)
    {
	privmsg(u->nick, BotServ.nick, "Syntax: \002DEL\002 <nickname>");
	return;
    }

    if ((bot = FindBot(parv[1])) != NULL)
    {
	DelBot(parv[1]);
	privmsg(u->nick, BotServ.nick, "Bot [\002%s\002] removed.", parv[1]);
	return;
    }

    privmsg(u->nick, BotServ.nick, "Bot [\002%s\002] not found.", parv[1]);
  return;
}

static int bs_change (struct User *u, int parc, char *parv[])
{
    privmsg(u->nick, BotServ.nick, "Under Construction");
    return;
}

static int bs_join(struct User *u, int parc, char *parv[])
{
    struct Bot *bot;
    char    channel[CHANLEN];

    if (parc < 3)
    {
	privmsg(u->nick, BotServ.nick, "Syntax: \002JOIN\002 <bot nick> <channel>");
	return;
    }

    /* chop the channel down to length */
    strlcpy (channel, parv[2], CHANLEN+1);

    if (!is_channel(channel))
    {
	privmsg(u->nick, BotServ.nick, "Invalid channel name: %s", channel);
	return;
    }

    /* XXX check for channel owner or valid access */

    if ((bot = FindBot(parv[1])) == NULL)
    {
	privmsg(u->nick, BotServ.nick, "Bot \002%s\002 does not exist.", parv[1]);
	return;
    }

    if (BotOnChannel(bot, channel) > 0)
    {
	privmsg(u->nick, BotServ.nick, "Bot \002%s\002 already in %s", parv[1], channel);
	return;
    }

    AddChannelToBot (bot, channel);

	ServerJoin (bot->nick, channel);
    sts(":%s MODE %s +o %s", me.name, channel, bot->nick);
}

static int bs_part (struct User *u, int parc, char *parv[])
{
    struct Bot *bot;
    char channel[CHANLEN+1];

    if (parc < 3)
    {
	privmsg(u->nick, BotServ.nick, "Syntax: \002PART\002 <bot nick> <channel>");
	return;
    }

    strlcpy (channel, parv[2], CHANLEN+1);

    if (!is_channel(channel))
    {
	privmsg(u->nick, BotServ.nick, "Invalid channel name: %s", channel);
	return;
    }

    /* XXX check for channel owner or valid access */

    if ((bot = FindBot(parv[1])) == NULL)
    {
	privmsg(u->nick, BotServ.nick, "Bot \002%s\002 does not exist.", parv[1]);
	return;
    }

    if (BotOnChannel(bot, channel) <= 0)
    {
	privmsg(u->nick, BotServ.nick, "Bot \002%s\002 not in %s",  bot->nick, channel);
	return;
    }

    RemoveChannelFromBot (bot, channel);

    sts(":%s PART %s :Leaving", bot->nick, channel);
}


/* FindBot - Find existing bot
**
*/
struct Bot *FindBot(char *nick)
{
    struct Bot *bot;
    dlink_node *dl;

    DLINK_FOREACH(dl, BotServ.bots.head)
    {
        bot = (struct Bot *)dl->data;
        if (!strcasecmp(nick, bot->nick))
            return bot;
    }
    return NULL;
}

/* DelBot - Remove a bot
**
*/
int DelBot (char *nick)
{
    struct Bot *bot;
    dlink_node *dl, *tdl;

    DLINK_FOREACH_SAFE(dl, tdl, BotServ.bots.head)
    {
        bot = (struct Bot *)dl->data;
        if (!strcasecmp(nick, bot->nick))
        {
            /* First we need to QUIT the bot. */
            sts (":%s QUIT :Bot Removed", bot->nick);
            dlink_delete(dl, &BotServ.bots);
            dlink_free(dl);
            free(bot);
            return 1;
        }
    }
    return 0;
}

/* AddBot - Adding a new bot
**
** return values:
** 0 or above = successful addition
** -1 = bot exists
*/
int AddBot (char *nick, char *user, char *host, char *name)
{
    struct Bot *bot;
    dlink_node *dl;

    if (bot = FindBot(nick))
        return -1;

    bot = calloc(1, sizeof(struct Bot));

    strlcpy (bot->nick, nick, sizeof(bot->nick));
    strlcpy (bot->user, user, sizeof(bot->user));
    strlcpy (bot->host, host, sizeof(bot->host));
    strlcpy (bot->name, name, sizeof(bot->name));
    bot->modes |= UMODE_INVIS;

    dl = dlink_create ();
    dlink_add (bot, dl, &BotServ.bots);

    /* Bot added to our list, now to intoduce him */
    /*sts ("NICK %s 1 1 +i %s %s %s %s :%s", bot->nick, bot->user, bot->host, me.name, bot->host, bot->name);*/
	sts ("NICK %s 1 1 +i %s %s %s :%s", bot->nick, bot->user, bot->host, me.name, bot->name);

    return 0;
}

int BotOnChannel (struct Bot *bot, char *channel)
{
    dlink_node *dl;
    char *chan;

    DLINK_FOREACH(dl, bot->channels.head)
    {
	chan = (char *)dl->data;
	if (!strcasecmp(chan, channel))
	    return 1;
    }
    return 0;
}

void AddChannelToBot (struct Bot *bot, char *channel)
{
    dlink_node *dl;
    dl = dlink_create ();
    dlink_add (strdup(channel), dl, &bot->channels);
    return;
}

void RemoveChannelFromBot (struct Bot *bot, char *channel)
{
    dlink_node *dl, *tdl;
    char *chan;

    DLINK_FOREACH_SAFE(dl, tdl, bot->channels.head)
    {
	chan = (char *)dl->data;
	if (!strcasecmp(chan, channel))
	{
	    dlink_delete(dl, &bot->channels);
	    dlink_free (dl);
	    free (chan);
	    return;
	}
    }
}

void IntroduceBots (void)
{
	struct Bot *bot;
	dlink_node *dl;
	
	DLINK_FOREACH (dl, BotServ.bots.head)
	{
		bot = (struct Bot *)dl->data;
		 sts ("NICK %s 1 1 +i %s %s %s :%s", bot->nick, bot->user, bot->host, me.name, bot->name);
	}

	return;
}

void BSJoinChannels (void)
{
	struct Bot *bot;
	char	*channel;
	dlink_node	*dl, *dl2;

	DLINK_FOREACH (dl, BotServ.bots.head)
	{
		bot = (struct Bot *)dl->data;
		DLINK_FOREACH (dl2, bot->channels.head)
		{
			channel = (char *)dl2->data;
			ServerJoin(bot->nick, channel);
			sts(":%s MODE %s +o %s", me.name, channel, bot->nick);
		}
	}
}


#endif /* defined(NICKSERVICES) && defined(BOTSERVICES) */
