#include "fwserv.h"
#include "config.h"
#include "chanserv.h"
#include "channel.h"
#include "user.h"

#ifdef USE_MYSQL
  #include <mysql.h>
  #include "_mysql.h"
#endif

#define MTYPE_NULL	0
#define MTYPE_ADD	1
#define MTYPE_DEL	2

struct CMode
{
  char	mode;
  unsigned int	value;
};

static struct CMode mode_list[] =
{
  { 'i',	CMODE_INVITE },
  { 'm',	CMODE_MOD },
  { 'n',	CMODE_NOEXT },
  { 't',	CMODE_TOPIC },
  { 'p',	CMODE_PRIV },
  { 'a',	CMODE_ANON },
  { 'O',	CMODE_OPERONLY },
  { 'A',	CMODE_ADMINONLY },
  { 'N',	CMODE_NETADMINONLY },
  { 's',	CMODE_SEC },
  { '\0',	0L }
};

static struct CMode status_mode_list[] =
{
 { 'o',		CMODE_OP },
 { 'h',		CMODE_HALFOP },
 { 'v',		CMODE_VOICE },
 { '\0',	0L }
};

/* XXX NEW FUNCTION */
void AddChannel (char *name, long ts)
{
	struct Channel *channel;
	dlink_node *dl;

	channel = calloc (1, sizeof(struct Channel));
	strlcpy (channel->name, name, sizeof(channel->name));
	channel->ts = ts;

	dl = dlink_create ();
	dlink_add (channel, dl, &channel_list);

	log (L_INFO, "Creating channel %s", name);

	return;
}
/* XXX NEW FUNCTION */
struct Channel *FindChannel (char *name)
{
	struct Channel *channel;
	dlink_node *dl;

	DLINK_FOREACH (dl, channel_list.head)
	{
		channel = (struct Channel *)dl->data;
		if (!strcasecmp(channel->name, name))
			return channel;
	}

	return NULL;
}

/* XXX NEW FUNCTION */
void RemoveUserFromChannel (struct Channel *channel, char *nick)
{
	dlink_node *dl, *tdl;
	struct ChanUser *cuser;

	if (!IsOnChan(channel, nick))
		return;

	DLINK_FOREACH_SAFE (dl, tdl, channel->users.head)
	{
		cuser = (struct ChanUser *)dl->data;
		if (!strcasecmp(cuser->nick, nick))
		{
			dlink_delete (dl, &channel->users);
			dlink_free (dl);
			free (cuser);
			return;
		}
	}

	channel->usercount--;
}

void RemoveChannel (char *channel)
{
	dlink_node *dl, *tdl;
	struct Channel *chan;

	DLINK_FOREACH_SAFE (dl, tdl, channel_list.head)
	{
		chan = (struct Channel *)dl->data;
		if (!strcasecmp(chan->name, channel))
		{
			dlink_delete (dl, &channel_list);
			dlink_free (dl);
			free (chan);
			return;
		}
	}

	return;
}

void ChannelMode (struct Channel *c, int parc, char *parv[])
{
  int i, matched, parpos = 1, what = MTYPE_ADD;
  int access = 0;
  char *pos = parv[0];
  struct User *u;
  struct RChannel *rchan;
  struct ChanUser *cuser;
  dlink_node *dl;

  if ((pos == NULL) || (*pos == '\0'))
    return;

  if (*pos == '0')
    return;

  for (; *pos != '\0'; pos++)
  {
    matched = FALSE;

    if (*pos == '+')
    {
      what = MTYPE_ADD;
      continue;
    }
    if (*pos == '-')
    {
      what = MTYPE_DEL;
      continue;
    }

	rchan = FindRChannel(c->name);

    for (i=0; mode_list[i].mode != '\0'; i++)
    {
      if (*pos == mode_list[i].mode)
      {
        matched = TRUE;
        if (what == MTYPE_ADD)
          c->modes |= mode_list[i].value;
        else
          c->modes &= ~mode_list[i].value;
        break;
      }
    }

    if (matched)
      continue;

    for (i=0; status_mode_list[i].mode != '\0'; i++)
    {
      if (*pos == status_mode_list[i].mode)
      {
		u = FindUser (parv[parpos]);
		cuser = FindChanUser (c, parv[parpos]);

        switch (*pos)
        {
          case 'o':
            if (what == MTYPE_ADD)
			{
				access = GetAccessLevel(rchan, u->nick);
				if ((rchan) && (IsSecure(rchan))) {
					if (!IsIdent(u)) {
						sts (":%s MODE %s -o %s", ChanServ.nick, rchan->name, u->nick);
						log (L_INFO, "%s not identified, not oping", u->nick);
					}

					else if ((access < rchan->levels->autoop) && !IsFounder(rchan, u->nick))
					{
						sts (":%s MODE %s -o %s", ChanServ.nick, rchan->name, u->nick);
						log (L_INFO, "Unprivleged user %s", u->nick);
					}

					else {
						log (L_INFO, "OPing %s on %s", u->nick, c->name);
						cuser->modes |= CMODE_OP;
					}
				}

				else if ((rchan) && (access < rchan->levels->autoop))
				{
					log (L_INFO, "Unprivleged user %s", u->nick);
				}
				else {
					log (L_INFO, "OPing %s on %s", u->nick, c->name);
					cuser->modes |= CMODE_OP;
				}

              log(1, "Oping %s", u->nick);

			}
            else
			{
              log(1, "Deoping %s", u->nick);
				cuser->modes &= ~CMODE_OP;
			}
            break;

          case 'v':
            if (what == MTYPE_ADD)
			{
				cuser->modes |= CMODE_VOICE;
              log(1, "Voicing %s", u->nick);
			}
            else
			{
				cuser->modes &= ~CMODE_VOICE;
              log(1, "Devoicing %s", u->nick);
			}
            break;

          case 'h':
            if (what == MTYPE_ADD)
			{
				cuser->modes |= CMODE_HALFOP;
              log(1, "Halfoping %s", u->nick);
			}
            else
			{
				cuser->modes &= ~CMODE_HALFOP;
              log(1, "Dehalfoping %s", u->nick);
			}
            break;
        }
        matched = TRUE;
        parpos++;
        continue;
      }
    }

    if (matched)
      continue;
  }

  return;
}

void
AddUserToChannel (struct Channel *c, char *user)
{
  int modifiers;
  dlink_node *dl;
	struct	RChannel *rchan;
	struct  User	*u;
	struct	ChanUser	*cuser;
	int level = 0;


  while (*user != '\0')
  {
    if (*user == '@')
    {
      modifiers |= CMODE_OP;
      user++;
      continue;
    }
    if (*user == '+')
    {
      modifiers |= CMODE_VOICE;
      user++;
      continue;
    }
    if (*user == '%')
    {
      modifiers |= CMODE_HALFOP;
      user++;
      continue;
    }

    break;
  }

	u = FindUser (user);
	rchan = FindRChannel (c->name);

	log ("Adding %s to channel %s", user, c->name);

	/* We need to allocate the channel user before doing anything */
	cuser = calloc(1, sizeof(struct ChanUser));
	strlcpy (cuser->nick, user, sizeof(cuser->nick));

	if (modifiers & CMODE_OP)
	{
		if (rchan != NULL)
		{
			level = GetAccessLevel (rchan, user);

			if (level < rchan->levels->autoop)
				sts (":%s MODE %s -o %s", ChanServ.nick, rchan->name, user);
			else
				cuser->modes |= CMODE_OP;
		}
		else
		{
			cuser->modes |= CMODE_OP;
		}
	}

	if (modifiers & CMODE_HALFOP)
	{
		cuser->modes |= CMODE_HALFOP;
	}

	if (modifiers & CMODE_VOICE)
	{
		cuser->modes |= CMODE_VOICE;
	}

	if (!(IsChanOp(cuser)) && !(IsChanHOp(cuser)) && !(IsChanVoice(cuser)))
	{
		if ((rchan) && (IsIdent(u)))
		{
			level = GetAccessLevel (rchan, user);

			if (level >= rchan->levels->autoop)
			{
				sts (":%s MODE %s +o %s", ChanServ.nick, rchan->name, user);
				cuser->modes |= CMODE_OP;
			}

			if ((level >= rchan->levels->autohop) && (level < rchan->levels->autoop))
			{
				sts (":%s MODE %s +o %s", ChanServ.nick, rchan->name, user);
				cuser->modes |= CMODE_HALFOP;
			}

			if ((level >= rchan->levels->autovoice) && (level < rchan->levels->autohop))
			{
				sts (":%s MODE %s +v %s", ChanServ.nick, rchan->name, user);
				cuser->modes |= CMODE_VOICE;
			}
		}

  }

	dl = dlink_create ();
	dlink_add (cuser, dl, &c->users);
	c->usercount++;
}

int IsOnChan (struct Channel *c, char *user)
{
	struct ChanUser *cuser;
	dlink_node *dl;

	DLINK_FOREACH (dl, c->users.head)
	{
		cuser = (struct ChanUser *)dl->data;
		if (!strcasecmp(user, cuser->nick))
			return 1;
	}

	return 0;
}

struct ChanUser *FindChanUser (struct Channel *channel, char *nick)
{
	dlink_node *dl;
	struct ChanUser *cuser;

	DLINK_FOREACH (dl, channel->users.head)
	{
		cuser = (struct ChanUser *)dl->data;
		if (!strcasecmp(cuser->nick, nick))
			return cuser;
	}

	return NULL;
}
