#include "fwserv.h"
#include "nickserv.h" /* nickreg */
#include "chanserv.h"
#include "memoserv.h"
#include "botserv.h"
#include "data.h"

#ifdef USE_MYSQL
  MYSQL *conn;
#endif

int LoadNicks (void);
int LoadChannels (void);
void SaveNicks (void);

void SaveMemos (void);
void LoadMemos (void);

void SaveChannels (void);

int ClearTables (void);

#ifdef USE_MYSQL
int LoadNicks (void)
{
    MYSQL_RES *res;
    MYSQL_ROW row;

    struct Nick *n;
    dlink_node *dl;
    int count = 0, ret = 0;

    res = sql_query_res ("SELECT * FROM %s", NickServ.table);

	if (res == NULL)
		return -2;

    if (mysql_num_fields(res) != 10) {
		log (1, "Incorrect amount of fields in table");
		return -2;
    }

  while ((row = mysql_fetch_row(res)) != NULL)
  {
    n = calloc(1, sizeof(struct Nick));

    if (row[9] == NULL) {
	log (1, "No nickname associated with data");
	ret = -1;
	free (n);
	continue;
    }

    strcpy(n->nick, row[0]);
    strcpy(n->password, row[1]);
    strcpy(n->last_quit, row[4]);
    strcpy(n->email, row[6]);
    strcpy(n->url, row[7]);
    strcpy(n->last_address, row[8]);

    n->flags = atoi(row[2]);
    n->last_seen = atoi(row[3]);
    n->registered = atoi(row[5]);
    n->expire = atoi(row[9]);

    count++;

    dl = dlink_create ();
    dlink_add(n, dl, &nick_list);
  }

  mysql_free_result(res);

  log(1, "Loaded %d Nicknames", count);

  return (ret);
}

int LoadChannels (void)
{
    MYSQL_RES	*res, *res2;
    MYSQL_ROW	row, row2;
	struct RChannel *rchan;
    struct Nick *n;
	struct AccessEntry *aentry;
	struct Levels *level;

    dlink_node *dl, *adl;

    int count = 0;

    res = sql_query_res ("SELECT * FROM %s", ChanServ.table);

  while ((row = mysql_fetch_row(res)) != NULL)
  { 
    if ((n = FindNick(row[4])) == NULL)
    {
      log(FATAL, "FATAL: Founder %s not registered!", row[4]);
    }

    rchan = calloc(1, sizeof(struct RChannel));

    strlcpy(rchan->name, row[0], CHANLEN);
    if (row[1] != NULL) 
      strlcpy(rchan->topic, row[1], TOPICLEN);
    strlcpy(rchan->pass, row[2], MAXPASS);
    if (row[3] != NULL)
      strlcpy(rchan->desc, row[3], BUFSIZE);

    rchan->founder = n;
	rchan->flags = atoi(row[5]);

	res2 = sql_query_res ("SELECT * FROM %s WHERE channel = '%s'", ChanServ.access_table, rchan->name);

	if (res2)
	{
		while ((row2 = mysql_fetch_row(res2)) != NULL)
		{
			aentry = calloc (1, sizeof(struct AccessEntry));
			log (L_INFO, "Adding access entry (Channel: %s) (Nickname: %s) (Level: %s)", row2[0], row2[1], row2[2]);
			if (row2[1] != NULL)
				strlcpy(aentry->nick, row2[1], sizeof(aentry->nick));
			else
			{
				log (L_INFO, "%s %s %s", row2[0], row2[1], row2[2]);
				free(aentry);
				continue;
			}
			aentry->level = atoi(row2[2]);

			adl = dlink_create ();
			dlink_add (aentry, adl, &rchan->access);
		}
	}

	mysql_free_result (res2);

	res2 = sql_query_res ("SELECT * FROM %s WHERE channel = '%s'", "ChannelLevels", rchan->name);

	if (res2)
		row2 = mysql_fetch_row (res2);

	if (row2)
	{
		level = calloc(1, sizeof (struct Levels));

		level->autodeop = atoi(row2[1]);
		level->autovoice = atoi(row2[2]);
		level->autoop = atoi(row2[3]);
		level->cmdvoice = atoi(row2[4]);
		level->cmdop = atoi(row2[5]);
		level->cmdaccess = atoi(row2[6]);
		level->cmdinvite = atoi(row2[7]);
		level->autohop = atoi(row2[8]);
		level->cmdhop = atoi(row2[9]);
		level->cmdunban = atoi(row2[10]);
		level->cmdakick = atoi(row2[11]);
		level->cmdclear = atoi(row2[12]);
		level->cmdset = atoi(row2[13]);
		level->superop = atoi(row2[14]);

		rchan->levels = level;
	}

	dl = dlink_create();
    dlink_add(rchan, dl, &rchannel_list);
    count++;
  }
  
  mysql_free_result(res);
  mysql_free_result(res2);
  
  log(1, "Loaded %d Channels", count);
	return 0;
}

void SaveNicks (void)
{
  struct Nick *n;
  dlink_node *dl;

  char query[BUFSIZE];

  log (1, "Saving Nicknames");

  conn = connect_db();

  DLINK_FOREACH(dl, nick_list.head)
  {
    n = (struct Nick *)dl->data;
    n->flags &= ~NS_IDENTIFIED;
    sprintf(query, "INSERT INTO %s VALUES ('%s', '%s', %d, %d, '%s', %d, '%s', '%s', '%s', %d)", NickServ.table, n->nick, n->password, n->flags, n->last_seen, n->last_quit, n->registered,
	n->email, n->url, n->last_address, n->expire);

    if (mysql_query (conn, query) < 0)
      log (FATAL, "FATAL: mysql_query failed: %s", mysql_error(conn));
  }

  mysql_close(conn);
  return;
}

void SaveChannels (void)
{
  struct RChannel *rchan;
  struct AccessEntry *aentry;
  dlink_node *dl, *adl;

  DLINK_FOREACH(dl, rchannel_list.head)
  {
	rchan = (struct RChannel *)dl->data;
	if (!sql_query("INSERT INTO %s VALUES ('%s', '%s', '%s', '%s', '%s', %d)", ChanServ.table, rchan->name, rchan->topic, rchan->pass, rchan->desc, rchan->founder->nick, rchan->flags))
	{
		log (L_WARNING, "Error saving channels");
		return;
	}
	DLINK_FOREACH(adl, rchan->access.head)
	{
		aentry = (struct AccessEntry *)adl->data;
		if (!sql_query("INSERT INTO %s VALUES ('%s', '%s', %d)", ChanServ.access_table, rchan->name, aentry->nick, aentry->level))
		{
			log (L_WARNING, "Error saving access list entry for channel %s", rchan->name);
			continue;
		}
	}
	if (!sql_query("INSERT INTO %s VALUES ('%s', %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d)", "ChannelLevels", rchan->name, rchan->levels->autodeop, rchan->levels->autovoice, rchan->levels->autoop, rchan->levels->cmdvoice, rchan->levels->cmdop, rchan->levels->cmdaccess, rchan->levels->cmdinvite, rchan->levels->autohop, rchan->levels->cmdhop, rchan->levels->cmdunban, rchan->levels->cmdakick, rchan->levels->cmdclear, rchan->levels->cmdset, rchan->levels->superop))
	{
		log (L_WARNING, "Error saving level list entry for channel %s", rchan->name);
	}
  }

  return;
}

void SaveMemos (void)
{
  struct Memo *memo;
  dlink_node *dl;
  char query[BUFSIZE];

  conn = connect_db();

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    sprintf(query, "INSERT INTO %s VALUES ('%s', '%s', '%s', %d, %d, %d)", MemoServ.table, memo->to, memo->from, memo->memo, memo->time_sent, memo->flags, memo->id);

    if (mysql_query (conn, query) < 0)
      log (FATAL, "FATAL: mysql_query failed: %s", mysql_error(conn));
  }

  mysql_close(conn);
}

void SaveBots (void)
{
	struct Bot *bot;
	char	*channel;
	dlink_node *dl, *dl2;

	DLINK_FOREACH (dl, BotServ.bots.head)
	{
		bot = (struct Bot *)dl->data;
		if (!sql_query("INSERT INTO %s VALUES ('%s', '%s', '%s', '%s')", BotServ.table, bot->nick, bot->user, bot->host, bot->name))
		{
			log (L_WARNING, "Error saving bot entry %s", bot->nick);
			continue;
		}
		DLINK_FOREACH (dl2, bot->channels.head)
		{
			channel = (char *)dl2->data;
			if (!sql_query("INSERT INTO %s VALUES ('%s', '%s')", BotServ.channel_table, bot->nick, channel))
			{
				log (L_WARNING, "Error saving channel entry for bot %s", bot->nick);
				continue;
			}
		}
	}
}

int LoadBots (void)
{
	MYSQL_RES	*res, *res2;
	MYSQL_ROW	row, row2;
	dlink_node	*dl, *dl2;
	struct Bot *bot;

	res = sql_query_res ("SELECT * FROM %s", BotServ.table);

	while ((row = mysql_fetch_row(res)) != NULL)
	{
		bot = calloc(1, sizeof(struct Bot));
		strlcpy (bot->nick, row[0], sizeof(bot->nick));
		strlcpy (bot->user, row[1], sizeof(bot->user));
		strlcpy (bot->host, row[2], sizeof(bot->host));
		strlcpy (bot->name, row[3], sizeof(bot->name));

		res2 = sql_query_res ("SELECT * FROM %s WHERE botnick = '%s'", BotServ.channel_table, bot->nick);

		while ((row2 = mysql_fetch_row(res2)) != NULL)
		{
			dl2 = dlink_create ();
			dlink_add (strdup(row2[1]), dl2, &bot->channels);
		}

		dl = dlink_create ();
		dlink_add (bot, dl, &BotServ.bots);
	}
}

void LoadMemos (void)
{
  MYSQL_RES *res;
  MYSQL_ROW row;

  struct Memo *memo;
  dlink_node *dl;
  char query[BUFSIZE];
  int count = 0;

  res = sql_query_res ("SELECT * FROM %s", MemoServ.table);

  while ((row = mysql_fetch_row(res)) != NULL)
  {
    dl = dlink_create();
    memo = calloc(1, sizeof(struct Memo));

    strcpy(memo->to, row[0]);
    strcpy(memo->from, row[1]);
    strcpy(memo->memo, row[2]);

    memo->time_sent = atoi(row[3]);
    memo->flags = atoi(row[4]);
    memo->id = atoi(row[5]);

    /*debug("D: %s %s %s %d %d %d", row[0], row[1], row[2], atoi(row[3]), atoi(row[4]), atoi(row[5]));*/

    dlink_add_tail(memo, dl, &memo_list);
    count++;
  }

  mysql_free_result(res);

  log(1, "Loaded %d Memos", count);
}

#endif

char* make_salt()
{
        static char saltChars[] =
          "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
        static char salt[3];
 
        salt[0] = saltChars[random() % 64];
        salt[1] = saltChars[random() % 64];
        salt[2] = 0;
                        
        return salt;
}


void
ReloadNicks (void)
{
	dlink_node *dl,*tdl;
	struct Nick *nick;

  DLINK_FOREACH(dl, nick_list.head)
  {
    dlink_delete(dl, &nick_list);
  }

  LoadNicks();
  return;
}

void ReloadMemos (void)
{
  dlink_node *dl;

  DLINK_FOREACH(dl, memo_list.head)
    dlink_delete(dl, &memo_list);

  LoadMemos();
  return;
}

void DoReload (void)
{
	dlink_node *dl, *tdl;
	struct Nick *nick;
	struct RChannel *rchan;

	DLINK_FOREACH (dl, nick_list.head)
	{
		nick = (struct Nick *)dl->data;
		nick->flags |= NS_DELETE;
	}

	DLINK_FOREACH (dl, rchannel_list.head)
	{
		rchan = (struct RChannel *)dl->data;
		rchan->flags |= CS_DELETE;
	}

	if (LoadNicks() == -2)
	{
		wallops(OperServ.nick, "Error loading databases, using old entries.");
		DLINK_FOREACH_SAFE (dl, tdl, nick_list.head)
		{
			nick = (struct Nick *)dl->data;
			if (!nick->flags & NS_DELETE)
			{
				dlink_delete(dl, &nick_list);
				dlink_free(dl);
				free(nick);
				continue;
			}
			else
			{
				nick->flags &= ~NS_DELETE;
			}
		}
	}

	if (LoadChannels() == -2)
	{
		wallops(OperServ.nick, "Error loading databases, using old entries.");
		DLINK_FOREACH_SAFE (dl, tdl, rchannel_list.head)
		{
			rchan = (struct RChannel *)dl->data;
			if (!rchan->flags & CS_DELETE)
			{
				dlink_delete(dl, &rchannel_list);
				dlink_free(dl);
				free(rchan);
				continue;
			}
			else
			{
				rchan->flags &= ~CS_DELETE;
			}
		}
	}

	DLINK_FOREACH_SAFE (dl, tdl, nick_list.head)
	{
		nick = (struct Nick *)dl->data;
		if (nick->flags & NS_DELETE)
		{
			dlink_delete(dl, &nick_list);
			dlink_free(dl);
			free(nick);
			continue;
		}
	}

	DLINK_FOREACH_SAFE (dl, tdl, rchannel_list.head)
	{
		rchan = (struct RChannel *)dl->data;
		if (rchan->flags & CS_DELETE)
		{
			dlink_delete(dl, &rchannel_list);
			dlink_free(dl);
			free(rchan);
			continue;
		}
	}
		
}


#ifdef USE_MYSQL
int ClearTables (void)
{

#ifdef NICKSERVICES
    if (!sql_query("DELETE FROM %s", NickServ.table))
		return -1;

#ifdef CHANNELSERVICES
    if (!sql_query("DELETE FROM %s", ChanServ.table))
		return -1;

	if (!sql_query("DELETE FROM %s", ChanServ.access_table))
		return -1;

	if (!sql_query("DELETE FROM %s", "ChannelLevels"))
		return -1;
#endif

#ifdef MEMOSERVICES
    if (!sql_query("DELETE FROM %s", MemoServ.table))
		return -1;
#endif

#ifdef BOTSERVICES
    if (!sql_query("DELETE FROM %s", BotServ.table))
		return -1;

	if (!sql_query("DELETE FROM %s", BotServ.channel_table))
		return -1;
#endif
#endif /* NICKSERVICES */

    return 0;
}

void DoSave (void)
{
    int ret;
    log (L_INFO, "Saving Databases...");

    if (ClearTables() < 0)
    {
	log (1, "Error clearing MySQL tables. Save failed.");
	return;
    }

#ifdef NICKSERVICES
    SaveNicks ();

#ifdef CHANNELSERVICES
    SaveChannels ();
#endif

#ifdef MEMOSERVICES
    SaveMemos ();
#endif

#ifdef BOTSERVICES
	SaveBots ();
#endif
#endif /* NICKSERVICES */
	log (L_INFO, "Databases saved.");
}
#endif /* USE_MYSQL */
