#include "fwserv.h"
#include "memoserv.h"
#include "nickserv.h"
#include "utils.h"
#include "config.h"

void StartInitialEvents(void);
static void evh_SaveDatabases(void *unused);
static void evh_PurgeMemos(void *unused);
static void evh_Routine(void *unused);

void StartInitialEvents(void)
{
  eventAdd("Save Databases", evh_SaveDatabases, NULL, 1800);
  eventAdd("Purge Memos", evh_PurgeMemos, NULL, 900);
  eventAdd("Routine", evh_Routine, NULL, 60);
}

static void evh_SaveDatabases(void *unused)
{
    DoSave ();
  return;
}

static void evh_PurgeMemos(void *unused)
{
void PurgeMemos (void)
{
  dlink_list purge_list;
  dlink_node *dl, *tdl, *tdl_;
  struct Memo *memo, *tmemo;
  struct MemoPurge *memop;
  int id, count = 0;

  purge_list.head = purge_list.tail = NULL;

  DLINK_FOREACH_SAFE(dl, tdl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (memo->flags & MS_MARKED)
    {
      dlink_delete(dl, &memo_list);
      dlink_free(dl);
      free(memo);
    } else {
      tdl_ = dlink_create();
      memop = calloc(1, sizeof(struct MemoPurge));

      strcpy(memop->to, memo->to);
      strcpy(memop->from, memo->from);
      strcpy(memop->memo, memo->memo);

      memop->time_sent = memo->time_sent;
      memop->flags = memo->flags;

      dlink_add(memop, tdl_, &purge_list);
    }
  }

  FlushMemos();

  DLINK_FOREACH(dl, purge_list.head)
  {
    memop = (struct MemoPurge *)dl->data;
    id = 0;
    DLINK_FOREACH_PREV(tdl, memo_list.tail)
    {
      memo = (struct Memo *)tdl->data;
      if (!strcasecmp(memo->to, memop->to)) {
        id = memo->id + 1;
        break;
      }
    }

    if (id == 0) id = 1;

    tdl_ = dlink_create();
    tmemo = calloc(1, sizeof (struct MemoPurge));

    strcpy(tmemo->to, memop->to);
    strcpy(tmemo->from, memop->from);
    strcpy(tmemo->memo, memop->memo);

    tmemo->time_sent = memop->time_sent;
    tmemo->flags = memop->flags;
    tmemo->id = id;

    dlink_add_tail(tmemo, tdl_, &memo_list);
    count++;
  }

  sts("WALLOPS :Purged %d memos", count);

}
}

static void evh_Routine (void *unused)
{
  dlink_node *dl;
  struct Nick *nick;
  struct Nick *n;
  time_t now = time(NULL);

  DLINK_FOREACH(dl, newnick_list.head)
  {
    n = (struct Nick *)dl->data;
    if (n->flags & NS_NOEXPIRE)
      continue;

    if ((now - n->last_seen) > NickServ.expire)
    {
		wallops(OperServ.nick, "Nick %s expired [%s]", n->nick, timediff(time(NULL) - n->last_seen));
      DelNick(n->nick);
    }
  }
}

void evh_reconnect_uplink(void *unused)
{
  ConnectUplink(me.uplink, me.port);
  eventDelete(evh_reconnect_uplink, NULL);
}
