#include "config.h"
#include "fwserv.h"
#include "signals.h"
#include "nickserv.h"
#include "operserv.h"
#include "chanserv.h"
#include "user.h"
#include "memoserv.h"
#include "server.h"
#include "channel.h"

void init_lists (void);
void init_logs (void);
void do_loop (void);
void log(int, char *, ...);
void debug(char *, ...);
void make_daemon(void);
void ShowListeners(void);
void ConnectUplink(char *, int);

void show_uptime (char *origin);

void close_listensocks (void);
void disconnect_users (void);

void write_pidfile(int);
void check_pidfile(void);

int uplinksock;
int allow_scanning;

static int Server_recieveFrom(int, char *);

static int c_ping (char *, int, char **);
static int c_version (char *, int, char **);
static int c_privmsg (char *, int, char **);
static int c_nick (char *, int, char **);
static int c_quit (char *, int, char **);
#ifdef IRCD_HYBSVS
static int c_svshost (char *, int, char **);
#endif
static int c_scan (char *, int, char **);
static int c_sjoin (char *, int, char **);
static int c_eob (char *, int, char **);
static int c_stats (char *, int, char **);
static int c_admin (char *, int, char **);
static int c_server (char *, int, char **);
static int c_mode (char *, int, char **);
static int c_squit (char *, int, char **);
static int c_topic (char *, int, char **);
static int c_part (char *, int, char **);

static struct CmdTable cmd_table[] =
{
  { "PING",     NULL,   c_ping,		NULL },
  { "VERSION",  NULL,   c_version,	NULL },
  { "PRIVMSG",  NULL,   c_privmsg,	NULL },
  { "NICK",     NULL,   c_nick,		NULL },
  { "QUIT",     NULL,   c_quit,		NULL },
#ifdef IRCD_HYBSVS
  { "SVSHOST",  NULL,   c_svshost,	NULL },
#endif
  { "SJOIN",    NULL,   c_sjoin,	NULL },   
  { "EOB",      NULL,   c_eob,		NULL },
  { "STATS",	NULL,	c_stats,	NULL },
  { "ADMIN",	NULL,	c_admin,	NULL },
  { "SERVER",	NULL,	c_server,	NULL },
  { "MODE",	NULL,	c_mode,		NULL },
  { "SQUIT",	NULL,	c_squit,	NULL },
  { "TOPIC",	NULL,	c_topic,	NULL },
  { "PART",		NULL,	c_part,		NULL },
  { (char *)0,  NULL,   (int (*)()) 0,	(int)0 }
};

static int
c_squit (char *origin, int parc, char *parv[])
{
  del_server(parv[0]);
  return 0;
}

static int
c_part (char *origin, int parc, char *parv[])
{
	struct Channel *c;
	struct ChanUser *cuser;

	if (origin)
	{
		if (parc < 1)
		{
			return 0;
		}

		if ((c = FindChannel (parv[0])) != NULL)
		{
			if ((cuser = FindChanUser (c, origin)) != NULL)
			{
				RemoveUserFromChannel (c, origin);
				return 0;
			}
		}
	}
	return 0;
}

static int c_topic (char *origin, int parc, char *parv[])
{
	struct RChannel *rchan;

	if (origin)
	{
		if (parc < 2)
		{
			log (L_INFO, "Not enough parameters sent with TOPIC");
			return 0;
		}

		rchan = FindRChannel (parv[0]);
		if (!rchan)
			return 0;

		if (IsTopicLock(rchan))
		{
			SetTopic(rchan);
			return 0;
		}
		return 0;
	}

	log (L_INFO, "TOPIC without an origin");
	return 0;
}

static int
c_admin (char *origin, int parc, char *parv[])
{
  struct User *u;
  if (origin)
  {
    u = FindUser(origin);
    if (u == NULL)
    {
      log(1, "c_admin(): Unknown User %s", origin);
      return;
    }

    sts(":%s %d %s :Administrative info about %s", me.name, 256, origin, me.name);
    sts(":%s %d %s :%s", me.name, 257, origin, me.desc);
    sts(":%s %d %s :%s", me.name, 258, origin, me.adminname);
    sts(":%s %d %s :%s", me.name, 259, origin, me.adminemail);
  }
}

static int
c_mode (char *origin, int parc, char *parv[])
{

    if (origin != NULL)
    {
	if (is_channel(parv[0]))
	    ChannelMode(FindChannel(parv[0]), parc - 1, &parv[1]);
	else
	    UserMode(FindUser(parv[0]), parc - 1, &parv[1]);
	return 0;
    }

    log(1, "MODE without origin.");
    return 1;
}

static int
c_server (char *origin, int parc, char *parv[])
{
  struct Server *server;

  if (parc < 3)
  {
    log(1, "Missing parameters in SERVER");
    return;
  }

  if (origin)
  {
    if ((server = find_server(parv[0])) == NULL)
      add_server(parv[0], parv[2]);

    return;
  }

  if ((server = find_server(parv[0])) == NULL)
    add_server(parv[0], parv[2]);

  SendNicks();
  JoinChannels();
  OSJoinChannels();
  IntroduceBots();
  BSJoinChannels();
#ifdef SPY
  JoinSpy();
#endif
  return;
}

static int
c_stats (char *origin, int parc, char *parv[])
{
  struct User *u;
  if (origin)
  {
    u = FindUser(origin);
    if (u == NULL)
    {
      log(1, "c_stats(): Unknown User: %s", origin);
      return;
    }

    if (parc < 1)
    {
      return;
    }

    switch(*parv[0])
    {
      case 'C':
        break;

      case 'E':
      case 'e':
        show_events(u->nick);
        break;

      case 'U':
      case 'u':
        show_uptime(u->nick);
        break;
    }
  }

}

#ifdef IRCD_HYBSVS
static int
c_svshost (char *origin, int parc, char *parv[])
{
  struct User *u;
  log(1, "%s", parv[0]);
  if (origin)
  {
    if (parc != 3)
    {
      log(1, "Not enough params in SVSHOST %d", parc);
      return;
    }
    if ((u = FindUser(origin)) == NULL)
    {
      log(1, "Nick %s Not Found", parv[0]);
      return;
    }
    log(1, "Changing Hostname");
    strlcpy(u->hostname, parv[2], sizeof(u->hostname));
    return;
  }
  else
  {
    log(1, "No Origin with SVSHOST");
    return;
  }
}
#endif

static int
c_eob (char *origin, int parc, char *parv[])
{
	if (origin)
	{
		if (!me.silent)
			wallops(me.name, "Uplink Complete. %s", timediff(time(NULL) - me.burststart));
		log(1, "Uplink Established");

		allow_scanning = 1;
		/*ScanAll();*/
		return 0;
	}
	log (L_INFO, "EOB without an origin.");
	return 0;
}

static int
c_sjoin (char *origin, int parc, char *parv[])
{
  /*struct NewChannel *c;*/
  struct Channel *c;

  char *modev[3];
  char *userv[SJPARA];

  int userc;
  int i;
  int modec = 1;

  if (origin != NULL)
  {
    modev[0] = parv[2];

    if (parc > 4)
    {
      modev[modec] = parv[3];
      modec++;
    }

    if (parc > 5)
    {
      modev[modec] = parv[4];
      modec++;
    }

    c = FindChannel(parv[1]);
    if (c == NULL)
    {
      AddChannel(parv[1], atol(parv[0]));
      c = FindChannel(parv[1]);
    }

    ChannelMode(c, modec, modev);

    userc = sjtoken(parv[parc - 1], userv);

    for (i=0; i < userc; i++)
      AddUserToChannel(c, userv[i]);

    return 0;
  }

  log(1, "SJOIN without origin.");
  return 1;
}

static int
c_quit (char *origin, int parc, char *parv[])
{
  struct User *u;
  struct NewNick *n;

  if (origin)
  {
    if (parc < 1)
    {
      log(1, "Not enough params in QUIT. %d", parc);
      return;
    }

    if ((u = FindUser(origin)) == NULL)
    {
      log(1, "Unknown User: %s", origin);
      return;
    }

    if (u->n)
    {
      if (u->n->flags & NS_IDENTIFIED)
      {
        strcpy(u->n->last_quit, parv[0]);
        strcpy(u->n->last_address, u->hostname);
        u->n->last_seen = time(NULL);
        u->n->flags &= ~NS_IDENTIFIED;
      }
    }

    DelUser(origin);
    return 0;
  }

  log(1, "QUIT without origin.");
  return 1;
}

static int
c_nick (char *origin, int parc, char *parv[])
{
  struct User *u;
  struct Nick *n;

  if (origin)
  {
    if (parc < 2)
    {
      log(1, "c_nick(): Insufficient parcs (%d)", parc);
      return;
    }

    if ((u = FindUser(origin)) == NULL)
    {
      log(1, "Unknown User: %s", origin);
      return;
    }

    if (u->n)
    {
      if (u->n->flags & NS_IDENTIFIED)
      {
        u->n->flags &= ~NS_IDENTIFIED;
        u->n->last_seen = time(NULL);
        strlcpy(u->n->last_address, u->hostname, HOSTLEN);
      }
      u->n = NULL;
    }

    u->ts = atoi(parv[1]);

    strlcpy(u->nick, parv[0], NICKLEN);

    if ((n = FindNick(parv[0])) != NULL)
    {
      if (IsLinked(origin, parv[0]) > 0)
      {
        u->n = n;
        u->n->flags |= NS_IDENTIFIED;
      } else {
        u->n = n;
        u->flags &= ~LVL_IDENT|LVL_ADMIN;

        if (u->n)
        {
          privmsg(u->nick, NickServ.nick, RPL_NEED_IDENTIFY);
          privmsg(u->nick, NickServ.nick, RPL_NEED_IDENTIFY2);
        }
      }
    }
    return;
  } else {
    if ((parc == 9) || (parc == 8))
      AddUser (parv[0], atoi(parv[2]), parv[3], parv[4], parv[5], parv[6], (parc > 8)? parv[7] : NULL, parv[(parc < 9)? 7 : 8]);
    return;
  }
}

static int
c_ping (char *origin, int parc, char *parv[])
{
  if(origin)
  {
    if (parc < 2)
    {
      log(1, "Not enough parcs");
      return;
    }
    sts(":%s PONG %s :%s", me.name, parv[1], parv[0]);
    return;
  }
  else
    sts("PONG :%s", parv[0]);
}

static int
c_privmsg (char *origin, int parc, char *parv[])
{
    struct User *u;

    if (origin != NULL)
    {
	if (parc < 2)
	{
	    log(1, "c_privmsg(): Missing parameter in PRIVMSG");
	    return;
	}

	if ((u = FindUser(origin)) == NULL)
	{
	    log(1, "c_privmsg(): Unknown User %s", origin);
	    return;
	}

	/*if (!strcasecmp(parv[0], n_UtilServ))
	    utilserv(u, parv[0], parv[1]);*/
	if (!strcasecmp(parv[0], ChanServ.nick))
	    chanserv(u, parv[0], parv[1]);
	else if (!strcasecmp(parv[0], OperServ.nick))
	    operserv(u, parv[0], parv[1]);
	else if (!strcasecmp(parv[0], NickServ.nick))
	    nickserv(u, parv[0], parv[1]);
	else if (!strcasecmp(parv[0], BotServ.nick))
	    botserv(u, parv[0], parv[1]);
	else if (!strcasecmp(parv[0], MemoServ.nick))
	    memoserv(u, parv[0], parv[1]);

	return;
    }

    log (1, "c_privmsg(): No origin");
    return;
}

c_version (char *origin, int parc, char *parv[])
{
  if(origin)
  {
    sts(":%s %d %s :%s", me.name, 351, origin, version_string);
    return;
  }
  else
    log(1, "No origin with VERSION");
}

void
init_lists (void)
{

  log(1, "Initializing Lists");
  connection_list.head = services_list.head = service_checks.head = nick_list.head = NULL;
  connection_list.tail = services_list.tail = service_checks.tail = nick_list.tail = NULL;

  channel_list.head = ourchannel_list.head = nickreg_list.head = NULL;
  channel_list.tail = ourchannel_list.tail = nickreg_list.tail = NULL;

  memo_list.head = nickkill_list.head = server_list.head = newuser_list.head = NULL;
  memo_list.tail = nickkill_list.tail = server_list.tail = newuser_list.tail = NULL;

  newnick_list.head = user_list.head = linked_nicks.head = NULL;
  newnick_list.tail = user_list.tail = linked_nicks.tail = NULL;

/*
  scanned_list.head = NULL;
  scanned_list.tail = NULL;
*/

  return;
}

void
init_logs (void)
{
  FILE *fp;

  if(!(fp = fopen(LPATH, "w")))
    {
      fprintf(stderr, "Error Opening %s", LPATH);
      return;
    }

  if(fprintf(fp, ""))
    ;

  fclose(fp);

#ifdef IO_LOG
  if(!(fp = fopen(ILOG, "w")))
  {
    fprintf(stderr, "Error Opening %s", ILOG);
    return;
  }

  fprintf(fp, "");
  fclose(fp);
#endif
}

void
do_loop (void)
{
  int SelectResult;
  struct timeval TimeOut;
  fd_set readfds, writefds, nullfds;

  struct Connection *conn;
  struct NickKill *nickkill;
  struct User *u;
  static char buf[BUFSIZE];

  dlink_node *dl, *tdl;

  TimeOut.tv_sec = 1;
  TimeOut.tv_usec = 0;

  while (Running)
  {
    memset(buf, '\0', BUFSIZE);

    eventRun();

    DLINK_FOREACH_SAFE(dl, tdl, nickkill_list.head)
    {
      nickkill = (struct NickKill *)dl->data;
      u = FindUser (nickkill->nick);
      if ((time(NULL) - nickkill->since) == 30)
        privmsg(u->nick, NickServ.nick, "You now have 30 seconds");
      if ((time(NULL) - nickkill->since) > 60)
      {
        sts(":%s KILL %s :Nick Enforcement", me.name, nickkill->nick);
        dlink_delete(dl, &nickkill_list);
      }
    }

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&nullfds);

    DLINK_FOREACH(dl, connection_list.head)
    {
      conn = (struct Connection *)dl->data;

      if(conn->flags & C_CONNECTING)
      {
        if(conn->flags & C_UPLINK)
          FD_SET(conn->fd, &writefds);
        if(conn->flags & C_SERVICE)
          FD_SET(conn->fd, &writefds);
      }
      else
      {
        FD_SET(conn->fd, &readfds);
      }
    }

    SelectResult = select(FD_SETSIZE, &readfds, &writefds, &nullfds, &TimeOut);

    if(SelectResult > 0)
    {
      DLINK_FOREACH(dl, connection_list.head)
      {
        conn = (struct Connection *)dl->data;

        if(FD_ISSET(conn->fd, &readfds))
        {
          if(conn->flags & C_LISTEN)
          {
            struct Connection *c;
            struct sockaddr_in *incoming_addr;

            dlink_node *dl;

            char desc[BUFSIZE];

            int asock;
                  
            int addrlen = sizeof(struct sockaddr);

            incoming_addr = calloc(1, sizeof(struct sockaddr_in));
            asock = accept(conn->fd, (struct sockaddr *)incoming_addr, (socklen_t *)&addrlen);

            if (asock < 0)
            {
              listen(conn->fd, 5);
              free(incoming_addr);
              continue;
            }

            else
            {
              dl = dlink_create();
              c = calloc(1, sizeof(struct Connection));

              c->fd = asock;
              c->flags |= C_USER;

              sprintf(desc, "Connected User From: %s", inet_ntoa(incoming_addr->sin_addr));
              strlcpy(c->desc, desc, sizeof(c->desc));

              dlink_add(c, dl, &connection_list);

              log(1, "User Connected From %s [fd %d] (%s)", inet_ntoa(incoming_addr->sin_addr), asock, desc);

              listen(conn->fd, 5);
            }
            continue;
          }
          /* Connected socket */
          process_read(conn);
          continue;
        }
        else if(FD_SET(conn->fd, &writefds))
        {
          if (conn->flags & C_CONNECTING)
          {
            if(conn->flags & C_UPLINK)
            {
              int ret = sts("PASS %s :TS", me.pass);
              if (ret == 1)
              {
                log(1, "Uplink Lost");
                del_connection(conn->fd);
                uplinksock = -1;
                eventAdd("Reconnect Uplink", evh_reconnect_uplink, NULL, me.reconnect);
                continue;
              }
              sts("CAPAB :%s", gen_capabs());
              /*sts("SERVER %s 1 :%s%s", me.name, (me.hidden > 0)? "(H) " : "", me.desc);*/
              sts("SERVER %s 1 :%s%s", me.name, (me.hidden > 0)? "(H) " : "", me.desc);
              sts("SVINFO 5 3 0 :%ld", time(NULL));
			  me.burststart = time(NULL);

              conn->flags &= ~C_CONNECTING;

              continue;
            }
            else if (conn->flags & C_SERVICE)
            {
              int ret = write(conn->fd, "\r\n", 4);

              if (ret == -1)
              {
                log(1, "Service: %s [Down]", conn->desc);
                close(conn->fd);
                del_connection(conn->fd);
                continue;
              }

              log(1, "Service: %s [Up]", conn->desc);
              close(conn->fd);
              del_connection(conn->fd);
              continue;
            }
          }
        }
      }
    }
    else
    {
      if(SelectResult == 0)
        ;
      else if((SelectResult == -1) && (errno == EINTR))
        ;
      else
        log(1, "Unknown Error");
    }
  }
}

void
log(int level, char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];
  FILE *fp;
       
  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  strlcat(buf, "\n", BUFSIZE);

if (IsNoFork())
  printf(buf);

  if(!(fp = fopen(LPATH, "a")))
    {
      fprintf(stderr, "Error Opening %s", LPATH);
      return;
    }

  fprintf(fp, "%s", buf);

  fclose(fp);

  va_end(ap);

  if (level > me.loglevel)
	return;

  if (level == FATAL)
  {
    printf("%s", buf);
    exit(-1);
  }

  return;
}

void debug (char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];

if (!IsDebug())
  return;

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);
  log(1, "%s", buf);
  va_end(ap);
}

#ifdef IO_LOG
void
io_log(char *fmt, ...)
{
  FILE *fp;
  va_list ap;
  char buf[BUFSIZE];

  va_start(ap, fmt);
  vsnprintf(buf, BUFSIZE, fmt, ap);

  strlcat(buf, "\n", BUFSIZE);

  if(IsNoFork())
    printf(buf);

  fp = fopen(ILOG, "a");

  fprintf(fp, "%s", buf);

  fclose(fp);

  va_end(ap);
  return;
}
#endif /* IO_LOG */

void
make_daemon (void)
{
  int pid;
  if ((pid = fork()) < 0)
  {
    perror("fork");
    exit(EXIT_FAILURE);
  }
  else if (pid > 0)
  {
    write_pidfile(pid);
    me.pid = pid;
    printf("FWServ: starting...\n");
    printf("FWServ: pid %d\n", pid);
    exit(EXIT_SUCCESS);
  }
}

void
write_pidfile (int pid)
{
  FILE *fp;
  if(!(fp = fopen(PPATH, "w")))
  {
    log(1, "Couldn't write pidfile: %s", PPATH);
    return;
  }

  fprintf(fp, "%d", pid);
  fclose(fp);

  return;
}

void
check_pidfile (void)
{
  FILE *fp;
  char buff[BUFSIZE];
  pid_t pidfromfile;

  if (!(fp = fopen(PPATH, "r")))
  {
    return;
  }

  if(!(fgets(buff, 20, fp)))
  {
    return;
  }

  pidfromfile = atoi(buff);
  if (!kill(pidfromfile, 0))
  {
    printf("FWServ: daemon already running\n");
    exit(-1);
  }
}

void
ShowListeners(void)
{
  dlink_node *dl;
  struct Connection *conn;

  fprintf(stdout, "Showing Listeners\n");

  DLINK_FOREACH(dl, connection_list.head)
    {
      conn = (struct Connection *)dl->data;
      fprintf(stdout, "FD: %d\tDesc: %s\n", conn->fd, conn->desc);
    }
}

/* Connect Uplink
 * Some code copyright W.Campbell
 */
void
ConnectUplink(char *host, int port)
{
  struct hostent *hp;
  struct sockaddr_in sa;
  struct in_addr * in;
  int s;
  int optval;
  int flags;

  struct Connection *conn;
  dlink_node *dl;

  debug("D: Connecting Uplink %s %d", host, port);

  if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      log(1, "Couldn't Setup Uplink Socket");
      return;
    }

    if ((hp = gethostbyname (host)) == NULL)
    {
            log(1, "Cannot gethostbyname(%s)", host);
            return;
    }

	memset((void *) &sa, '\0', sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (port);
	memcpy((void *) &sa.sin_addr, (void *) hp->h_addr, hp->h_length);

	flags = fcntl(s, F_GETFL, 0);
	flags |= O_NONBLOCK;
	(void) fcntl(s, F_SETFL, flags);

  connect (s, (struct sockaddr *) &sa, sizeof (sa));

  dl = dlink_create();
  conn = calloc(1, sizeof(struct Connection));
  conn->fd = s;
  uplinksock = s;
  conn->flags = C_UPLINK|C_CONNECTING;
  strlcpy(conn->desc, "Services Uplink", sizeof(conn->desc));
  dlink_add(conn, dl, &connection_list);
}

void
close_listensocks (void)
{
  dlink_node *dl;
  struct Connection *c;

  DLINK_FOREACH(dl, connection_list.head)
  {
    c = (struct Connection *)dl->data;
    if (c->flags & C_LISTEN)
    {
      close(c->fd);
      del_connection(c->fd);
    }
  }
}

void
disconnect_users (void)
{
  dlink_node *dl;
  struct Connection *c;

  DLINK_FOREACH(dl, connection_list.head)
  {
    c = (struct Connection *)dl->data;
    if (c->flags & C_USER)
    {
      close(c->fd);
      del_connection(c->fd);
    }
  }
}

void
parse (char *line)
{
  char *origin = NULL;
  char *command = NULL;
  char *message = NULL;
  char *parv[BUFSIZE];
  char *pos;

  int parc, i;

  for (i = 0; i < BUFSIZE; i++)
    parv[i] = NULL;

#ifdef IO_LOG
  io_log("R: %s", line);
#endif

  if ((pos = strchr(line, ' ')))
  {
    *pos = '\0';
    pos++;

    if (*line == ':')
    {
      origin = line;
      origin++;

      if ((message = strchr(pos, ' ')))
      {
        *message = '\0';
        message++;
        command = pos;
      }
      else
      {
        command = pos;
        message = NULL;
      }
    }
    else
    {
      message = pos;
      command = line;
    }
  }

  if (message)
  {
    if (*message == ':')
    {
      message++;
      parv[0] = message;
      parc = 1;
    }
    else
    {
      parc = tokenize(message, parv);
    }
  }
  else
    parc = 0;

  for(i=0; cmd_table[i].handler; i++)
    {
      if(strcasecmp(command, cmd_table[i].name) == 0)
        {
          int status = cmd_table[i].handler(origin, parc, parv);
          return;
        }
    }
    

}

int tokenize(char *message, char **parv)
{
        char *pos = NULL; 
        char *next;
        int count = 0;
                        
        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }
                 
        /* First we find out if there's a : in the message, save that string
        ** somewhere so we can set it to the last param in parv
        ** Also make sure there's a space before it...if not, then we're
        ** screwed
        */
        pos = message;
        while (TRUE)
        {
                if ((pos = strchr(pos, ':')))
                {
                        pos--;
                        if (*pos != ' ')
                        {
                                /* There SHOULDN'T be a problem with this... */
                                pos += 2;
                                continue;
                        }
                        *pos = '\0';
                        /* This is BAD - get rid of the space before the : */
                        pos++;
                        *pos = '\0';
                        /* VERY BAD...over now though */
                        pos++;
                        break;
                }
                else
                {
                        break;
                }
        }

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */

        next = message;
        parv[0] = message;
        count = 1;

        while (*next)
        {
                if (count == MAXPARA - 1)
                {
                        /* We've reached one less than our max limit
                        ** to handle the parameter we already ripped off
                        */
                        return count;
                }
                if (*next == ' ')
                {
                        *next = '\0';
                        next++;
                        /* Eat any additional spaces */
                        while (*next == ' ')
                                next++;
                        /* If it's the end of the string, it's simply
                        ** an extra space before the :parameter.  Here we
                        ** break.
                        */
                        if (*next == '\0')
                                break;
                        parv[count] = next;
                        count++;
                }
                else
                {
                        next++;
                }
        }

        if (pos)
        {
                parv[count] = pos;
                count++;
        }

        return count;
}

/* A generic tokenizer, tokenizes based on a specified character */
int generictoken(char delim, int size, char *message, char **parv)
{
        char *next;
        int count;

        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */
        next = message;
        parv[0] = next;
        count = 1;

        while (*next)
        {
                /* This is fine here, since we don't have a :delimited
                ** parameter like tokenize
                */
                if (count == size)
                {
                    /* We've reached our limit */
                    return count;
                }
                if (*next == delim)
                {
                        *next = '\0';
                        next++;
                        /* Eat any additional delimiters */
                        while (*next == delim)
                                next++;
                        /* If it's the end of the string, it's simply
                        ** an extra space at the end.  Here we break.
                        */
                        if (*next == '\0')
                                break;
                        parv[count] = next;
                        count++;
                }
                else
                {
                        next++;  
                }
        }
                                
        return count;
}

void
show_uptime (char *origin)
{
  sts(":%s %d %s :*** Uptime: %s", me.name, 242, origin, timediff(time(NULL) - me.started));
  return;
}

void
TermUplink (void)
{
  uplinksock = -1;

  /*newnick_list.head = newnick_list.tail = NULL;
  InitServices();*/

  eventAdd("Reconnect Uplink", evh_reconnect_uplink, NULL, me.reconnect);
}
