#include "fwserv.h"
#include "config.h"
#include "signals.h"
#include "nickserv.h"
#include "operserv.h"
#include "server.h"

int Running;
int foreground = 0;
int dorehash = 0;
int allow_scanning = 0;
char *version_string;

static void CmdLine(int argc, char *argv[]);
static char *build_version(void);

int main(int argc, char *argv[])
{
  struct sigaction act;
  struct Server *server;
  act.sa_flags = 0;
  act.sa_handler = SIG_IGN;

  me.started = time(NULL);
  CmdLine(argc, argv);
     
  /* Before we do anything major see if we are already
     running once */
  /*check_pidfile();*/
     
  init_logs();
 
  if (chdir(DPATH))
  {
    log(1, "Couldn't chdir to %s [%d] (%s)", DPATH, errno, strerror(errno));
    exit(-1);
  }

  version_string = build_version();

  InitDefaults();
  ConfLoad();
  VerifyConf();
  
  /*NewDefaults();
  NewParseConf();*/
  init_lists();
  /*ReadOpers();*/
  eventInit();

  if (IsNoFork())
  {
    fprintf(stdout, "Starting FWServ");
    me.flags |= CMD_DEBUG;
  }
  else
    make_daemon();
  log(1, "Starting FWServ");
 
  setsid();

  signal(SIGTERM, SigTerm);
  signal(SIGHUP, SigHup);
  signal(SIGUSR1, SigUsr1);
  signal(SIGUSR2, SigUsr2);
  signal(SIGINT, SigInt);
  signal(SIGPIPE, SIG_IGN);

/*
  act.sa_handler = SigSegv;
  sigaddset(&act.sa_mask, SIGSEGV);
  sigaction(SIGSEGV, &act, 0);
*/

  add_server(me.name, me.desc);
  /*NewMe.server = find_server(NewSettings.sname);*/

  /*DoListen(16668);
  DoListen(16669);*/

#ifdef USE_MYSQL
    /* Connect our MySQL Server */
    me.mptr = connect_db();
#endif
     
  InitServices();

  LoadNicks();
  LoadMemos();
  LoadChannels();
  LoadBots();

  ConnectUplink(me.uplink, me.port);

  StartInitialEvents();
  Running = 1;

  do_loop();

  /* NOT REACHED */
  return 1;
}

static void
CmdLine (int argc, char *argv[])
{
  char c;

  while ((c=getopt(argc, argv, "dhn")) != -1)
  {
    switch(c)
    {
      case 'd':
        me.flags |= CMD_DEBUG;
        break;
      case 'n':
        me.flags |= CMD_NOFORK;
        break;
      case 'c':
        me.flags |= CMD_CHROOT;
        break;
    }
  }
  return;
}

static char
*build_version (void)
{
  static char version[BUFSIZE];
  version[0] = '\0';

  strcpy(version, PATCHLEVEL);
  strcat(version, "(");
  strcat(version, SERIALNUM);
  strcat(version, ")");

  return version;
}
