#include "fwserv.h"
#include "config.h"

#include "nickserv.h"
#include "memoserv.h"
#include "user.h"
#include <time.h>

#ifdef USE_MYSQL
  #include <mysql.h>
  #include "_mysql.h"
#endif

#if defined(NICKSERVICES) && defined(MEMOSERVICES)

void memoserv (struct User *, char *, char *);
int RecordMemo (struct Nick *, struct User *, char *);
char *MemoStatus (char *, int);
void PurgeMemos (void);
void FlushMemos (void);

void FlushMemoServTable ();

static int ms_help (struct User *, int, char **);

static int ms_send (struct User *, int, char **);
static int ms_read (struct User *, int, char **);
static int ms_list (struct User *, int, char **);

static int ms_mark (struct User *, int, char **);
static int ms_unmark (struct User *, int, char **);

static int ms_purge (struct User *, int, char **);

static struct CmdTable memoserv_table[] =
{
  { "HELP",	NULL,	ms_help,	LVL_NONE },

  { "SEND",	NULL,	ms_send,	LVL_IDENT },
  { "READ",	NULL,	ms_read,	LVL_IDENT },
  { "LIST",	NULL,	ms_list,	LVL_IDENT },

  { "MARK",	NULL,	ms_mark,	LVL_IDENT },
  { "UNMARK",	NULL,	ms_unmark,	LVL_IDENT },

  { "PURGE",	NULL,	ms_purge,	LVL_ADMIN },
  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0}
};

static int
ms_help (struct User *u, int parc, char *parv[])
{
  privmsg(u->nick, MemoServ.nick, "Not Available Yet");
  return;
}

static int
ms_send (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  struct Memo *memo;
  static char memobuf[BUFSIZE];
  dlink_node *dl;
  int i, memoid = 0;
  memobuf[0] = '\0';

  if (parc < 3)
  {
    privmsg(u->nick, MemoServ.nick, "Syntax: \002SEND\002 <nickname> <memo>");
    return;
  }

  if (!u->n)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, u->nick);
    return;
  }

  if ((n = FindNick(parv[1])) == NULL)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, parv[1]);
    return;
  }

  if (!(n->flags & NS_ALLOWMEMO))
  {
    privmsg(u->nick, MemoServ.nick, "Not allowed to memo %s", n->nick);
    return;
  }

  for(i=2; parv[i] != NULL; i++)
    if (memobuf[0] == '\0') {
      strcpy(memobuf, parv[i]);
    } else {
      strcat(memobuf, " ");
      strcat(memobuf, parv[i]);
    }

  RecordMemo(n, u, memobuf);
  return;
}

static int
ms_read (struct User *u, int parc, char *parv[])
{
  struct Nick *nick;
  struct Memo *memo;
  dlink_node *dl;

  if (parc < 2)
  {
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_READ_SYNTAX);
    return;
  }

  if (!u->n)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, u->nick);
    return;
  }

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(memo->to, u->nick))
      if (memo->id == atoi(parv[1]))
      {
	privmsg(u->nick, MemoServ.nick, RPL_MEMO_FROM, memo->from);
	privmsg(u->nick, MemoServ.nick, RPL_MEMO, memo->memo);
        memo->flags &= ~MS_NEW;
        return;
      }
  }
  return;
}

static int
ms_list (struct User *u, int parc, char *parv[])
{
  dlink_node *dl;
  struct Memo *memo;
  int count = 0;

  struct tm *ltm;
  static char fmtime[BUFSIZE];

  if (!u->n)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, u->nick);
    return;
  }

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(u->n->nick, memo->to))
    {
      ltm = localtime(&memo->time_sent);
      strftime(fmtime, BUFSIZE, FTMFMT, ltm);
      privmsg(u->nick, MemoServ.nick, RPL_MEMO_LIST, memo->id, MemoStatus(u->n->nick, memo->id), memo->from, fmtime);
      count++;
    }
  }
  if (count == 0)
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_NONE, u->nick);
  return;
}

static int
ms_mark (struct User *u, int parc, char *parv[])
{
  struct Nick *nick;
  struct Memo *memo;
  dlink_node *dl;

  if (parc < 2)
  {
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_MARK_SYNTAX);
    return;
  }

  if (!u->n)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, u->nick);
    return;
  }

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(memo->to, u->n->nick))
    {
      if (memo->id == atoi(parv[1]))
      {
        memo->flags |= MS_MARKED;
        memo->flags &= ~MS_NEW|MS_READ;
	privmsg(u->nick, MemoServ.nick, "Memo %s marked for deletion.", memo->id);
        return;
      }
    }
  }

  privmsg(u->nick, MemoServ.nick, RPL_MEMO_ILLEGAL_ID, atoi(parv[1]));
    
  return;
}

static int
ms_unmark (struct User *u, int parc, char *parv[])
{
  dlink_node *dl;
  struct Memo *memo;

  if (parc < 2)
  {
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_UNMARK_SYNTAX);
    return;
  }

  if (!u->nick)
  {
    privmsg(u->nick, MemoServ.nick, RPL_NOT_REGISTERED, u->nick);
    return;
  }

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(memo->to, u->n->nick))
    {
      if (memo->id == atoi(parv[1]))
      {
        memo->flags |= MS_MARKED;
        memo->flags &= ~MS_NEW|MS_READ;
	privmsg(u->nick, MemoServ.nick, "Memo %d no longer marked for deletion.", memo->id);
        return;
      }
    }
  }

  privmsg(u->nick, MemoServ.nick, RPL_MEMO_ILLEGAL_ID, atoi(parv[1]));
  return;
}

ms_purge (struct User *u, int parc, char *parv[])
{
  PurgeMemos();
  return;
}

void memoserv(struct User *u, char *target, char *param)
{
  char *parv[MAXPARA];
  int parc, i, status;
  static char buf[BUFSIZE];
  parc = 0;

  for (i = 0; i < MAXPARA; i++)
    parv[i] = NULL;

  strlcpy(buf, param, BUFSIZE);

  if (param != NULL)
    parc = msgtoken(param, parv);

  if (parc == 0)
    return;

  for (i = 0; memoserv_table[i].handler; i++)
  {

    if(!strcasecmp(parv[0], memoserv_table[i].name))
    {
      if ((memoserv_table[i].flag & LVL_IDENT) && !(u->n))
      {
        privmsg(u->nick, MemoServ.nick, RPL_MUST_IDENTIFY, memoserv_table[i].name);
        return;
      }

      if ((memoserv_table[i].flag & LVL_ADMIN) && !(u->flags & LVL_ADMIN))
      {
	privmsg (u->nick, MemoServ.nick, "Invalid Command: %s", parv[0]);
        return;
      }


      status = memoserv_table[i].handler(u, parc, parv);
      return;
    }
  }

    privmsg(u->nick, MemoServ.nick, "Invalid Command: %s", parv[0]);
    return;
}

int RecordMemo (struct Nick *n, struct User *u, char *memo)
{
  int memoid = 0;
  dlink_node *dl, *tdl;
  struct Memo *_memo;
  struct Memo *tmemo;

  struct User *user;

  DLINK_FOREACH_PREV(tdl, memo_list.tail)
  {
    tmemo = (struct Memo *)tdl->data;
    if (!strcasecmp(n->nick, tmemo->to))
    {
      memoid = tmemo->id + 1;
      debug("D: another memo found %d", tmemo->id);
      break;
    }
  }

  if (memoid == 0)
    memoid++;
  debug("D: Memo ID: %d", memoid);

  dl = dlink_create();
  _memo = calloc(1, sizeof(struct Memo));

  strcpy(_memo->to, n->nick);
  strcpy(_memo->from, u->n->nick);
  strcpy(_memo->memo, memo);
  _memo->time_sent = time(NULL);
  _memo->id = memoid;
  _memo->flags |= MS_NEW;

  dlink_add_tail(_memo, dl, &memo_list);

  privmsg(u->nick, MemoServ.nick, RPL_MEMO_SENT, n->nick);

  if ((user = FindUser(n->nick)) != NULL)
      if (user->n->flags & NS_IDENTIFIED)
	privmsg(user->nick, MemoServ.nick, RPL_MEMO_NEW, u->n->nick);

  return 0;
}

void CheckMemos (char *nick)
{
  struct Memo *memo;
  struct User *u;
  dlink_node *dl;
  int memos = 0;

  if ((u = FindUser(nick)) == NULL)
    return;

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(memo->to, u->n->nick) && (memo->flags & MS_NEW))
      memos++;
  }

  if (memos != 0)
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_SIGNON_NEW, memos);
  else
    privmsg(u->nick, MemoServ.nick, RPL_MEMO_SIGNON_NONE);
}

char *MemoStatus (char *nick, int id)
{
  dlink_node *dl;
  struct Memo *memo;

  static char status[BUFSIZE];
  strcpy(status, "R");

  DLINK_FOREACH(dl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (!strcasecmp(memo->to, nick))
      if (memo->id == id)
        if (memo->flags & MS_NEW)
          strcpy(status, "N");
        else if (memo->flags & MS_MARKED)
          strcpy(status, "D");
  }

  return status;
}

void FlushMemos (void)
{
  dlink_node *dl, *tdl;
  struct Memo *memo;

  DLINK_FOREACH_SAFE(dl, tdl, memo_list.head)
    dlink_delete(dl, &memo_list);
}

void PurgeMemos (void)
{
  dlink_list purge_list;
  dlink_node *dl, *tdl, *tdl_;
  struct Memo *memo, *tmemo;
  struct MemoPurge *memop;
  int id, count = 0;

  purge_list.head = purge_list.tail = NULL;

  DLINK_FOREACH_SAFE(dl, tdl, memo_list.head)
  {
    memo = (struct Memo *)dl->data;
    if (memo->flags & MS_MARKED)
      dlink_delete(dl, &memo_list);
    else {
      tdl_ = dlink_create();
      memop = calloc(1, sizeof(struct MemoPurge));

      strcpy(memop->to, memo->to);
      strcpy(memop->from, memo->from);
      strcpy(memop->memo, memo->memo);

      memop->time_sent = memo->time_sent;
      memop->flags = memo->flags;

      dlink_add(memop, tdl_, &purge_list);
    }
  }

  FlushMemos();

  DLINK_FOREACH(dl, purge_list.head)
  {
    memop = (struct MemoPurge *)dl->data;
    id = 0;
    DLINK_FOREACH_PREV(tdl, memo_list.tail)
    {
      memo = (struct Memo *)tdl->data;
      if (!strcasecmp(memo->to, memop->to)) {
        id = memo->id + 1;
        break;
      }
    }

    if (id == 0) id = 1;

    tdl_ = dlink_create();
    tmemo = calloc(1, sizeof (struct MemoPurge));

    strcpy(tmemo->to, memop->to);
    strcpy(tmemo->from, memop->from);
    strcpy(tmemo->memo, memop->memo);

    tmemo->time_sent = memop->time_sent;
    tmemo->flags = memop->flags;
    tmemo->id = id;

    dlink_add_tail(tmemo, tdl_, &memo_list);
    count++;
  }

  sts("WALLOPS :Purged %d memos", count);

}
#endif
