#include "fwserv.h"
#include "operserv.h"
#include "nickserv.h"
#include "utils.h"
#include "flags.h"
#include "user.h"
#include "data.h"

void operserv (struct User *, char *, char *);
void ServicesSpy (void);

struct Operator *FindOper(char *);
int FindOperHost (struct Operator *, struct User *);

static int os_identify (struct User *, int, char **);
static int os_die (struct User *, int, char **);
static int os_restart (struct User *, int, char **);
static int os_help (struct User *, int, char **);
static int os_join (struct User *, int, char **);
static int os_part (struct User *, int, char **);
static int os_rehash (struct User *, int, char **);
static int os_save (struct User *, int, char **);
static int os_reload (struct User *, int, char **);

static struct CmdTable operserv_table[] =
{
  { "HELP",	NULL,	os_help,	LVL_NONE },
  { "IDENTIFY",	NULL,	os_identify,	LVL_NONE },

  { "DIE",	NULL,	os_die,		LVL_ADMIN },
  { "SHUTDOWN",	NULL,	os_die,		LVL_ADMIN },
  { "RESTART",	NULL,	os_restart,	LVL_ADMIN },
  { "JOIN",	NULL,	os_join,	LVL_ADMIN },
  { "PART",	NULL,	os_part,	LVL_ADMIN },
  { "REHASH",	NULL,	os_rehash,	LVL_ADMIN },

  { "SAVE",	NULL,	os_save,	LVL_ADMIN },
  { "RELOAD",	NULL,	os_reload,	LVL_ADMIN },
  { (char *)0,	NULL,	(int (*)()) 0,	(int)0 }
};

static struct HelpTable help_table[] =
{
  { "LOGIN",	"help/operserv/login",	LVL_NONE },
  { "HELP",	"help/operserv/help",	LVL_NONE },
  { (char *)0,	(char *)0,		(int)0 }
};

struct Operator *FindOper (char *nick)
{
    struct Operator *op;
    dlink_node *dl;

    DLINK_FOREACH(dl, me.operators.head)
    {
	op = (struct Operator *)dl->data;
	if (!strcasecmp(op->nick, nick))
	    return op;
    }
    return NULL;
}

void AddOperServChannel (char *channel)
{
    char ochannel[CHANLEN+1];
    dlink_node *dl;

    strlcpy (ochannel, channel, sizeof(ochannel));

    dl = dlink_create();
    dlink_add (strdup(ochannel), dl, &OperServ.channels);
};

void OSJoinChannels (void)
{
    char *channel;
    dlink_node *dl;

    DLINK_FOREACH (dl, OperServ.channels.head)
    {
	channel = (char *)dl->data;
	sts(":%s SJOIN %ld %s +nt :@%s", me.name, time(NULL), channel, OperServ.nick);
	sts(":%s MODE %s +o %s", me.name, channel, OperServ.nick);
    }
}

void OSJoinChannel (char *channel)
{
    char ochannel[CHANLEN+1];
    strlcpy (ochannel, channel, sizeof(channel));
    sts(":%s SJOIN %ld %s +nt :@%s", me.name, time(NULL), channel, OperServ.nick);
    sts(":%s MODE %s +o %s", me.name, channel, OperServ.nick);
}

static int
os_save (struct User *u, int parc, char *parv[])
{
    DoSave ();
/*
  SaveNicks();
  SaveMemos();
  SaveChannels();
*/
  privmsg(u->nick, OperServ.nick, "Databases Saved");
  return;
}

static int
os_reload (struct User *u, int parc, char *parv[])
{
/*
  ReloadNicks();
  ReloadMemos();
*/
	DoReload();
  privmsg(u->nick, OperServ.nick, "Reloading Databases");
  return;
}

static int
os_rehash (struct User *u, int parc, char *parv[])
{
	wallops(OperServ.nick, "Rehashing, [Requested by: %s!%s@%s]", u->nick, u->username, u->hostname);
	privmsg(u->nick, OperServ.nick, "Rehashing...");
	log(1, "%s Rehashing config", u->nick);
	rehash();
	return;
}

static int
os_join (struct User *u, int parc, char *parv[])
{
  if (parc != 2)
  {
    notice(u, OperServ.nick, "Syntax: JOIN <#channel>");
    return;
  }

  notice(u, OperServ.nick, "Joining %s...", parv[1]);
  OSJoinChannel (parv[1]);
  return 0;
}

static int
os_part (struct User *u, int parc, char *parv[])
{
  if (parc != 2)
  {
    privmsg(u->nick, OperServ.nick, "Syntax: \002PART\002 <channel>");
    return;
  }

  privmsg(u->nick, OperServ.nick, "Parting %s...", parv[1]);
  sts(":%s PART %s :Leaving", OperServ.nick, parv[1]);
  return;
}

static int
os_help (struct User *u, int parc, char *parv[])
{
  int i;

  if (parc < 2)
  {
    send_help_file(u, OperServ.nick, "help/operserv/index", "index");
    return;
  }

  for (i=0; help_table[i].name; i++)
  {
    if (!strcasecmp(parv[1], help_table[i].name))
    {
      send_help_file(u, OperServ.nick, help_table[i].filename, parv[1]);
      return;
    }
  }
  privmsg(u->nick, OperServ.nick, "Ho help for %s", parv[1]);
  return;
}

static int
os_identify (struct User *u, int parc, char *parv[])
{
    struct Operator *op;
    char host[BUFSIZE];

    if (parc != 2)
    {
	privmsg(u->nick, OperServ.nick, "Syntax: IDENTIFY <password>");
	return;
    }

	if ((IsOSAdmin(u)))
		return;

    if ((OperServ.operonly > 0) && !(IsOper(u)))
    {
	privmsg (u->nick, OperServ.nick, "IRC Operators Only.");
#ifdef SPY
	Spy (OperServ.nick, "Failed IDENTIFY attempt from %s!%s@%s (Non Operator)", u->nick, u->username, u->hostname);
#endif
	return;
    }

    if ((op = FindOper(u->nick)) != NULL)
    {
	if (FindOperHost(op, u) < 1)
	{
	    notice (u, OperServ.nick, "No O-line");
#ifdef SPY
	    Spy (OperServ.nick, "Failed IDENTIFY attempt from %s!%s@%s (No O-line)", u->nick, u->username, u->hostname);
#endif
	    return;
	}

#ifdef CRYPT
	if (!strcmp(op->password, crypt(parv[1], op->password)))
#else
	if (!strcmp(op->password, parv[1]))
#endif
	{
	    notice(u, OperServ.nick, "You are now identified.");
	    if (HasOperFlag(op))
		u->flags |= LVL_OPER;
	    if (HasAdminFlag(op))
		u->flags |= LVL_ADMIN|LVL_OPER;

	    if ((u->flags & LVL_OPER) && !(u->flags & LVL_ADMIN))
			wallops (OperServ.nick, "%s is now a Services Operator", u->nick);
	    else
			wallops (OperServ.nick, "%s is now a Services Administrator", u->nick);

#ifdef SPY
	    Spy (OperServ.nick, "%s!%s@%s used IDENTIFY for %s", u->nick, u->username, u->hostname, u->nick);
#endif
	    return;
	}
	else
	{
#ifdef SPY
	    Spy (OperServ.nick, "Failed IDENTIFY attempt from %s!%s@%s (Incorrect Password)", u->nick, u->username, u->hostname);
#endif
	    notice(u, OperServ.nick, "Password Incorrect");
	    return;
	}
    }
    else
    {
#ifdef SPY
	Spy (OperServ.nick, "Failed IDENTIFY attempt from %s!%s@%s (No O-line)", u->nick, u->username, u->hostname);
#endif
	notice(u, OperServ.nick, "No O-line");
	return;
    }
}

static int
os_die (struct User *u, int parc, char *parv[])
{
  char reason[BUFSIZE];
  int i;

  reason[0] = '\0';

  if (parc < 2)
    sprintf(reason, "Shutdown Requested by %s", u->nick);
  else
	ExpandParv (reason, BUFSIZE, 1, parc, parv);

    DoSave ();
    privmsg(u->nick, OperServ.nick, "Shutting Down...");

	wallops (me.name, "Shutting Down [%s]", reason);
#ifdef SPY
    Spy (OperServ.nick, "%s!%s@%s used DIE [Reason: %s]", u->nick, u->username, u->hostname, reason);
#endif
    SendQuit (reason);

    log(1, "Shutting Down... %s [%s]", reason, u->nick);

    shutdown(uplinksock, SHUT_RDWR);
    del_connection(uplinksock);
    uplinksock = -1;
    mysql_close (me.mptr);
    sleep(1);
    exit(0);
}

static int
os_restart (struct User *u, int parc, char *parv[])
{
    char reason[BUFSIZE];
	reason[0] = '\0';

	if (parc < 2)
		sprintf (reason, "Restart Requested by %s", u->nick);
	else
		ExpandParv (reason, BUFSIZE, 1, parc, parv);

	wallops (me.name, "Restarting [%s]", reason);

    privmsg(u->nick, OperServ.nick, "Restarting...");
    SendQuit (reason);

    log(1, "%s [%s]", reason, u->nick);

	DoSave ();
    shutdown(uplinksock, SHUT_RDWR);
    del_connection(uplinksock);
    uplinksock = -1;
    mysql_close (me.mptr);
    sleep(1);
    execv(SPATH, NULL);

    /* NOT REACHED */
    return 0;
}

void operserv(struct User *u, char *target, char *param)
{
    char *parv[MAXPARA];
    int parc, i, status;
    static char buf[BUFSIZE];
    parc = 0;

	if ((OperServ.operonly > 0) && !(IsOper(u)))
	{
		privmsg(u->nick, OperServ.nick, "This service is for IRC Operators only.");
		return;
	}
  
    for (i = 0; i < MAXPARA; i++)
		parv[i] = NULL;
 
    strlcpy(buf, param, BUFSIZE);

#ifdef DEV_DEBUG
    log(1, "operserv(): u->nick: %s u->hostname: %s", u->nick, u->hostname);
    log(1, "operserv(): buf: %s target: %s", buf, target);
#endif

    if (param != NULL)
	parc = msgtoken(param, parv);
  
    if (parc == 0)
	return;
   
    for (i = 0; operserv_table[i].handler; i++)  
    {
	if(!strcasecmp(parv[0], operserv_table[i].name))
	{
	    if ((operserv_table[i].flag & LVL_ADMIN && !IsOSAdmin(u)))
	    {
			privmsg(u->nick, OperServ.nick, "Access Denied.");
			return;
	    }
	    if ((operserv_table[i].flag & LVL_OPER && !IsOSOper(u)))
	    {
			privmsg(u->nick, OperServ.nick, "Access Denied.");
			return;
	    }

	    status = operserv_table[i].handler(u, parc, parv);
	    return;
	}
    }

    privmsg(u->nick, OperServ.nick, "Invalid Command: %s", parv[0]);
    return;
}

int FindOperHost (struct Operator *op, struct User *u)
{
    dlink_node *dl;
    char *host, userhost[USERLEN+HOSTLEN+3];

    sprintf (userhost, "%s@%s", u->username, u->hostname);

    DLINK_FOREACH (dl, op->hosts.head)
    {
	host = (char *)dl->data;
	if (strchr(host, '@'))
	{
	    log (1, "Matching %s against %s", userhost, host);
	    if (match(host, userhost) != 1)
		return 1;
	}
	else
	{
	    log (1, "Matching %s against %s", u->hostname, host);
	    if (match(host, u->hostname) != 1)
		return 1;
	}
    }

    return 0;
}
