#include "config.h"
#include "fwserv.h"
#include "setup.h"

#ifdef PROXYSCAN
void ProxyScan(char *);
void open_proxy(OPM_T *, OPM_REMOTE_T *, int, void *);
void myend(OPM_T *, OPM_REMOTE_T *, int, void *);
void handle_error(OPM_T *scanner, OPM_REMOTE_T *remote, int err, void *data);
void ScanAll (void);
void DoScan (char *);

int complete = 0;
int in_burst;

#ifdef HAVE_GETADDRINFO
static struct addrinfo * gethostinfo(char const *host, int port);

void ProxyScan(char *ip)
{
  struct addrinfo *hostres;
  struct hostent *he;

  char buf[BUFSIZE];
  char *scanip;

  int j, host;

  int fd_limit, max_read;
  int timeout = 10;
  int scanport = 6668;
  fd_limit = 1024;
  max_read = 4096;

  strlcpy(buf, ip, BUFSIZE);
  j = strlen(buf);

  log(1, "Scanning: %s", ip);

  if (ip == NULL)
  {
    debug("D: ip == NULL");
    return;
  }

  if (scanip = strchr(ip, ':'))
    return;


  if (!isdigit(buf[j - 1]))
  {
    hostres = gethostinfo(ip, 0);
    if (hostres == NULL)
    {
      log(1, "gethostinfo(%s, %d) failed", ip, 0);
      return;
    }

    if (!(he = gethostbyname(ip))) {
      switch (h_errno) {
      case HOST_NOT_FOUND:
        sts("WALLOPS :Host '%s' is unknown, dummy.", ip);
        return;
      case NO_ADDRESS:
        sts("WALLOPS :The specified name '%s' exists, but has no address.", ip);
        return;
      case NO_RECOVERY:
        sts("WALLOPS :An unrecoverable error occured whilst resolving '%s'", ip);
        return;
      case TRY_AGAIN:
        sts("WALLOPS :A temporary nameserver error occured.");
        return;
      default:
        sts("WALLOPS :Unknown error resolving '%s'", ip);
        return;
      }

    }

    /*scanip = inet_ntoa(*((struct in_addr *) he->h_addr));*/
    /*inet_ntop( PF_UNSPEC, *he->h_addr, &scanip, 1023 );
    debug ("D: %s -", scanip);*/

    debug ("D: %s", inet_ntoa(*((struct in_addr *) he->h_addr)));

    if (scanip == NULL)
    {
      debug("D: scanip == NULL");
      return;
    }

    DoScan(scanip);
    return;
  }

  DoScan(ip);
}
#else
void ProxyScan(char *ip)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  struct in_addr * in;
  char buf[BUFSIZE];
  int j, host;

  strlcpy(buf, ip, BUFSIZE);
  j = strlen(buf);

  if (ip == NULL)
  {
    debug("D: ip == NULL");
    return;
  }

  if (!isdigit(buf[j - 1]))
  {
    if ((hp = gethostbyname(ip)) == NULL)
    {
      log(1, "gethostbyname(%s) failed, %s NOT scanned.", ip, ip);
      return;
    }
    memset((void *) &sa, '\0', sizeof(sa));
    sa.sin_family = AF_INET;
    sa.sin_port = NULL;
    memcpy((void *) &sa.sin_addr, (void *) hp->h_addr, hp->h_length);

    host = inet_ntoa(sa.sin_addr);

    if (host == NULL)
    {
      debug("D: host == NULL");
      return;
    }

    DoScan(host);
  } else {
    DoScan(ip);
  }
}
#endif

void DoScan (char *ip)
{
  OPM_T *scanner;
  OPM_REMOTE_T *remote_host;
  OPM_ERR_T err;

  struct Scanned *scanned;
  dlink_node *dl;

  int fd_limit, max_read;
  int timeout = 10;
  int scanport = 6668;
  fd_limit = 1024;
  max_read = 4096;

  /*DLINK_FOREACH(dl, scanned_list.head)
  {
    scanned = (struct Scanned *)dl->data;
    if (!strcasecmp(scanned->ip, ip))
      return;
  }*/

  log(1, "Scanning %s", ip);
  remote_host = opm_remote_create(ip);

  scanner = opm_create();
  
  opm_callback(scanner, OPM_CALLBACK_OPENPROXY, &open_proxy, NULL);
  opm_callback(scanner, OPM_CALLBACK_END, &myend, NULL);
  opm_callback(scanner, OPM_CALLBACK_ERROR, &handle_error, NULL);

  opm_config(scanner, OPM_CONFIG_FD_LIMIT, &fd_limit);
  opm_config(scanner, OPM_CONFIG_SCAN_IP, "24.81.119.139");
  opm_config(scanner, OPM_CONFIG_SCAN_PORT, &scanport);
  opm_config(scanner, OPM_CONFIG_MAX_READ, &max_read);
  opm_config(scanner, OPM_CONFIG_TIMEOUT, &timeout);
  /*opm_config(scanner, OPM_CONFIG_BIND_IP, "24.70.120.21");*/
  opm_config(scanner, OPM_CONFIG_TARGET_STRING, "*** Looking up your hostname...");
  opm_config(scanner, OPM_CONFIG_TARGET_STRING, "ERROR :Trying to reconnect too fast.");
  
  opm_addtype(scanner, OPM_TYPE_HTTP, 80);
  opm_addtype(scanner, OPM_TYPE_HTTP, 8080);
  opm_addtype(scanner, OPM_TYPE_HTTP, 3128);
  opm_addtype(scanner, OPM_TYPE_WINGATE, 23);
  opm_addtype(scanner, OPM_TYPE_ROUTER, 23);
  opm_addtype(scanner, OPM_TYPE_SOCKS4, 1080);
  opm_addtype(scanner, OPM_TYPE_SOCKS5, 1080);
  opm_addtype(scanner, OPM_TYPE_HTTPPOST, 3128);

  switch(err = opm_scan(scanner, remote_host))
  {
    case OPM_SUCCESS:
      break;
    case OPM_ERR_BADADDR:
      log(1, "Bad Address");
      opm_free(scanner);
      opm_remote_free(remote_host);
      break;
    default:
      log(1, "Unknown Error: %d", err);
  }

  while(!complete)
    opm_cycle(scanner);
  
  complete = 0;
  opm_free(scanner);
}

void
open_proxy (OPM_T *scanner, OPM_REMOTE_T *remote, int notused, void *data)
{
  sts("WALLOPS :Open Proxy: %s", remote->ip);
  sts(":%s KLINE * 0 * %s :Open Proxy (Contact jeremy@cvs.freeworld.nu)", n_UtilServ, remote->ip);
  log(1, "Open Proxy: %s:%d", remote->ip, remote->port);
  opm_end(scanner, remote);
}      

void      
myend(OPM_T *scanner, OPM_REMOTE_T *remote, int notused, void *data)
{
/*  struct Scanned *scanned;
  dlink_node *dl;

  dl = dlink_create();
  scanned = calloc(1, sizeof(struct Scanned));

  strlcpy(scanned->ip, remote->ip, sizeof(scanned->ip));
  dlink_add(scanned, dl, &scanned_list);*/

  opm_remote_free(remote);
  log(1, "Scan Done");
  complete = 1;
}

void handle_error(OPM_T *scanner, OPM_REMOTE_T *remote, int err, void *data)
{                      
   switch(err)
   {  
      case OPM_ERR_MAX_READ:
         printf("Reached MAX READ on %s:%d\n", remote->ip, remote->port);
         break;
      case OPM_ERR_BIND:
         printf("Unable to bind for %s:%d\n", remote->ip, remote->port);
         break;
      case OPM_ERR_NOFD:
         printf("Unable to allocate file descriptor for %s:%d\n", remote->ip, remote->port);
         break;
      default:
         printf("Unknown error on %s:%d, err = %d\n", remote->ip, remote->port, err);
   }
}

void ScanAll (void)
{
  dlink_node *dl;
  struct User *u;

  DLINK_FOREACH(dl, nick_list.head)
  {
    u = (struct User *)dl->data;
    ProxyScan(u->realhost);
  }
  return;
}

#ifdef HAVE_GETADDRINFO
/* Stolen from FreeBSD's whois client, modified for Stats by W. Campbell */
static struct addrinfo * gethostinfo(char const *host, int port)
{
        struct addrinfo hints, *res;
        int error;
        char portbuf[6];
 
        memset(&hints, 0, sizeof(hints));
        hints.ai_flags = 0;
        hints.ai_family = AF_UNSPEC;
        hints.ai_socktype = SOCK_STREAM;
        snprintf(portbuf, 6, "%d", port);
        error = getaddrinfo(host, portbuf, &hints, &res);
        if (error) {
                log("gethostinfo error: %s: %s", host, gai_strerror(error));
                return (NULL);
        }
        return (res);
}
#endif
#endif /* PROXYSCAN */
