#include "fwserv.h"
#include "config.h"
#include "signals.h"
#include "nickserv.h"
#include "operserv.h"
#include "server.h"

struct Server
*find_server (char *server)
{
  struct Server *s;
  dlink_node *dl;

  DLINK_FOREACH(dl, server_list.head)
  {
    s = (struct Server *)dl->data;

    if (!strcasecmp(server, s->name))
      return s;
  }

  debug("D: server %s not found.", server);
  return NULL;
}

void
add_server (char *name, char *desc)
{
  struct Server *s;
  dlink_node *dl;

  if ((s = find_server (name)) != NULL)
    return;

  dl = dlink_create();
  s = calloc(1, sizeof(struct Server));

  strlcpy(s->name, name, HOSTLEN);
  strlcpy(s->desc, desc, BUFSIZE);

  dlink_add(s, dl, &server_list);

  debug("D: server %s added.", name);

  return;
}

void
del_server (char *name)
{
  struct Server *s;
  dlink_node *dl, *tdl;

  DLINK_FOREACH_SAFE(dl, tdl, server_list.head)
  {
    s = (struct Server *)dl->data;
    if (!strcasecmp(s->name, name))
    {
      dlink_delete(dl, &server_list);
      dlink_free(dl);
      free(s);
      return;
    }
  }
}
