#include "fwserv.h"

void AddCheck(char *, char *, unsigned int);
void DelCheck(char *);
void CheckServices(void);
void CheckService(char *host, int port, char *desc);

struct ServiceCheck *FindCheck (char *);

void
AddCheck (char *name, char *host, unsigned int port)
{
  dlink_node *dl;
  struct ServiceCheck *service;

  dl = dlink_create();
  service = calloc(1, sizeof(struct ServiceCheck));

  strlcpy(service->name, name, sizeof(service->name));
  strlcpy(service->host, host, sizeof(service->host));

  service->port = port;
  service->status = 0;

  dlink_add(service, dl, &service_checks);

  return;
}

void
DelCheck (char *name)
{
  dlink_node *dl;
  struct ServiceCheck *service;

  DLINK_FOREACH(dl, service_checks.head)
  {
    service = (struct ServiceCheck *)dl->data;
    if (service->name == name)
    {
      dlink_delete(dl, &service_checks);
      free(service);
      dlink_free(dl);
      return;
    }
  }

  log(1, "Service Check not removed. %s", name);
  return;
}

struct ServiceCheck *FindCheck(char *name)
{
  struct ServiceCheck *service;
  dlink_node *dl;

  DLINK_FOREACH(dl, service_checks.head)
  {
    service = (struct ServiceCheck *)dl->data;
    if ((strcasecmp(service->name, name)) == 0)
      return service;
  }
  log(1, "Check not found");
  return NULL;
}

void
CheckServices (void)
{
  struct ServiceCheck *service;
  dlink_node *dl;

  DLINK_FOREACH(dl, service_checks.head)
  {
    service = (struct ServiceCheck *)dl->data;
    CheckService(service->host, service->port, service->name);
  }
}

void
CheckService (char *host, int port, char *desc)
{
  struct hostent *hp;
  struct sockaddr_in sa;
  struct in_addr * in;
  int s;
  int optval;
  int flags;

  struct Connection *conn;
  dlink_node *dl;

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    log(1, "Couldn't Setup Check Socket");
    return;
  }

  if ((hp = gethostbyname (host)) == NULL)
  {
    log(1, "Cannot gethostbyname(%s)", host);
    return;
  }

  memset((void *) &sa, '\0', sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons (port);
  memcpy((void *) &sa.sin_addr, (void *) hp->h_addr, hp->h_length);

  flags = fcntl(s, F_GETFL, 0);
  flags |= O_NONBLOCK;
  (void) fcntl(s, F_SETFL, flags);

  connect (s, (struct sockaddr *) &sa, sizeof (sa));
  dl = dlink_create();
  conn = calloc(1, sizeof(struct Connection));
  conn->fd = s;
  conn->flags = C_SERVICE|C_CONNECTING;
  strlcpy(conn->desc, desc, sizeof(conn->desc));
  dlink_add(conn, dl, &connection_list);
}
