/* FWServ: Light Weight IRC Services.
 * sock.c
 */

#include "fwserv.h"

void DoListen(unsigned int);
int ports = 0;

int Server_receiveFrom(int, char *);
void Service_processPacket(int, char *, char *);
void process_read(struct Connection *);

void
DoListen(unsigned int port)
{
  struct sockaddr_in OurSocket;
  struct Connection *conn;
  int mysock, flags, option;
  dlink_node *dl;
  char error[BUFSIZE];
  char desc[BUFSIZE];

  error[0] = '\0';

  option = 1;

  if((mysock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    {
      close(mysock);
      sprintf(error, "Unable to create stream socket: %s", strerror(errno));
      log(1, error);
      return;
    }

  memset(&OurSocket, 0, sizeof(OurSocket));
  OurSocket.sin_family = AF_INET;
  OurSocket.sin_addr.s_addr = htonl(INADDR_ANY);
  OurSocket.sin_port = htons(port);

  if (bind(mysock, (struct sockaddr *)&OurSocket, sizeof(OurSocket)) < 0)
    {
      close(mysock);
      sprintf(error, "Unable to bind to port %d: %s", port, strerror(errno));
      log(1, error);
      return;
    }

  if (listen(mysock, 4) < 0)
    {
      close(mysock);
      sprintf(error, "Unable to listen on port %d: %s", port, strerror(errno));
      log(1, error);
      return;
    }

  if(setsockopt(mysock, SOL_SOCKET, SO_REUSEADDR, (char *)&option, sizeof(option)) < 0)
    {
      sprintf(error, "SetSocketOptions: setsockopt(SO_REUSEADDR) failed: %s", strerror(errno));
      log(1, error);
      return;
    }

  flags = fcntl(mysock, F_GETFL, 0);
  flags |= O_NONBLOCK;
  (void)fcntl(mysock, F_SETFL, flags);

  dl = dlink_create();

  conn = calloc(1, sizeof(struct Connection));
  conn->fd = mysock;
  conn->flags |= C_LISTEN;
  sprintf(desc, "Port %d", port);
  strlcpy(conn->desc, desc, sizeof(conn->desc));
  dlink_add(conn, dl, &connection_list);

  ports++;

  return;
}

/* The following function was lifted
 * from Hwy's Stats
 */
void process_read(struct Connection *conn)
{                           
        static char buf[BUFSIZE + 1];
 
        memset(buf, '\0', BUFSIZE + 1);

        if (Server_receiveFrom(conn->fd, buf) <= 0)
        {
                if (conn->fd == uplinksock)
                {
                  /* Lost our uplink */
                  log(1, "Lost connection to uplink. (%d, %d)", conn->fd, uplinksock);
                  close(uplinksock);
                  del_connection(conn->fd);
                  TermUplink();
                  uplinksock = -1;
                  eventAdd("Reconnect Uplink", evh_reconnect_uplink, NULL, me.reconnect);
                  return;
                }
                else
                {
                  /* We lost a client */
                  log(1, "Lost connection to user, (%d)", conn->fd);
                  close(conn->fd);
                  del_connection(conn->fd);
                  return;
                }
        }
        Service_processPacket(conn->fd, buf, conn->hbuffer);
}

/* These two functions were lifted from oswald's NoteServ
 * which were copied from Hwy's Stats
 */
int Server_receiveFrom(int fd, char *buffer)
{
    /* This mustn't be BUFSIZE!!! */
    /* unless buffer is BUFSIZE + 1 */
    int n = read(fd, buffer, BUFSIZE + 1);
    buffer[n] = '\0';
    return n;
}

void Service_processPacket(int fd, char *buffer, char *backing)
{
    char *ptr;
    char buffer2[BUFSIZE * 2];

    while ((ptr = strchr(buffer, '\n')))
    {
        *ptr = '\0';
        if (ptr == buffer)
        {
                /* We got a \n in a packet by itself.  Process backing */
                if (*backing != '\0')
                {
                        strlcpy(buffer2, backing, BUFSIZE * 2);
                        if (fd == uplinksock)
                                parse(buffer2);
                        else
                          /* We'll just IO LOG these for now */
#ifdef IO_LOG
                          io_log("R: %s", buffer2);
#else
						  return;
#endif
                }
                buffer++;
                continue;
        }
        if (*(ptr - 1) == '\r')
                *(ptr - 1) = '\0';
        if (*backing != '\0')
        {
                snprintf(buffer2, (BUFSIZE * 2), "%s%s", backing, buffer);
                *backing = '\0';
        }
        else
        {
                strlcpy(buffer2, buffer, BUFSIZE * 2);
        }
        if (fd == uplinksock)
                parse(buffer2);
        else
#ifdef IO_LOG
          io_log("R: %s", buffer2);
#else
		  return;
#endif
        buffer = ptr + 1;
    }
    if (*buffer)
    {
        int pos = strlen(buffer);
        if (buffer[pos - 1] == '\r')
        {
                buffer[pos - 1] = '\0';
                strlcpy(buffer2, buffer, BUFSIZE * 2);
                if (fd == uplinksock)
                        parse(buffer2);
                else
#ifdef IO_LOG
                  io_log("R: %s", buffer2);
#else
				  return;
#endif
                *backing = '\0';
        }
        else
        {
                strlcpy(backing, buffer, BUFSIZE * 2);
        }
    }
}
