#include "fwserv.h"
#include "config.h"
#include "user.h"
#include "server.h"

#include "nickserv.h"
#include "botserv.h"

#define MTYPE_NULL      0
#define MTYPE_ADD       1
#define MTYPE_DEL       2

struct UMode
{
  char  mode;
  unsigned int  value;
};

static struct UMode user_modes[] =
{
  { 'i',    UMODE_INVIS },
  { 'a',    UMODE_ADMIN },
  { 'o',    UMODE_OPER },
  { '\0',   0L }
};

void AddUser (char *nick, int ts, char *modes, char *username, char *hostname, char *server, char *realhost, char *gecos)
{
    struct User *u;
    /*struct NickKill *nickkill;*/
    dlink_node *dl;

    if ((u = FindUser(nick)) != NULL)
    {
	log(1, "Trying to add a user already present (%s?)", nick);
	return;
    }

#ifdef PROXYSCAN
    if (allow_scanning)
	ProxyScan(realhost);
#endif

    dl = dlink_create();
    u = calloc(1, sizeof(struct User));

    strlcpy(u->nick, nick, NICKLEN);
    strlcpy(u->hostname, hostname, HOSTLEN);
	if (realhost == NULL)
		strlcpy (u->realhost, hostname, HOSTLEN+1);
	else
		strlcpy(u->realhost, realhost, HOSTLEN);
    strlcpy(u->username, username, USERLEN);
    strlcpy(u->gecos, gecos, REALLEN);

    u->server = find_server(server);
    u->n = FindNick(u->nick);

    u->flags |= LVL_NONE;
    u->ts = ts;

    if (!u->server)
	log(1, "Unable to find server for user %s (%s?) Possible spoofed server", nick, server);

    if (!u->n)
	log(1, "Unable to find nick for user %s (%s?) Probably not registered.", nick, u->nick);
    else
    {
	notice(u, NickServ.nick, RPL_NEED_IDENTIFY);
	notice(u, NickServ.nick, RPL_NEED_IDENTIFY2);
    }

/*
    if (u->n->flags & NS_KILL)
    {
      notice(u, NickServ.nick, "You have 60 seconds.");
      dl = dlink_create();
      nickkill = calloc(1, sizeof(struct NickKill));
      nickkill->since = time(NULL);
      strcpy(nickkill->nick, u->n->nick);
      dlink_add(nickkill, dl, &nickkill_list);
    }
  }
*/

    debug("D: AddUser (%s, %d, %s, %s, %s, %s, %s, %s)", u->nick, u->ts, modes, u->username, u->hostname, server, u->realhost, u->gecos);

    /* Update usermodes */
    UserMode (u, 1, &modes);

    /* Finally add user to our list */
    dlink_add(u, dl, &globaluser_list);

    return;
}

struct User *FindUser (char *nick)
{
    struct User *u;
    dlink_node *dl;

    DLINK_FOREACH(dl, globaluser_list.head)
    {
	u = (struct User *)dl->data;
	if (!strcasecmp(u->nick, nick))
	    return u;
    }

    debug("D: User %s not found.", nick);

    return NULL;
}

void DelUser (char *nick)
{
    struct User *u;
    dlink_node *dl, *tdl;

    DLINK_FOREACH_SAFE(dl, tdl, globaluser_list.head)
    {
	u = (struct User *)dl->data;
	if (!strcasecmp(u->nick, nick))
	{
	    dlink_delete(dl, &globaluser_list);
	    dlink_free(dl);
	    free(u);
	    debug("D: DelUser (%s): Successful", nick);
	    return;
	}
    }

    debug("D: DelUser (%s): Failed", nick);

}

void UserMode (struct User *u, int parc, char *parv[])
{
  int i, matched, parpos = 1, what = MTYPE_ADD;
  char *pos = parv[0];

  if ((pos == NULL) || (*pos == '\0'))
    return;

  if (*pos == '0')
    return;

  for (; *pos != '\0'; pos++)
  {
    matched = FALSE;
    if (*pos == '+')
    {
      what = MTYPE_ADD;
      continue;
    }
    if (*pos == '-')
    {
      what = MTYPE_DEL;
      continue;
    }

    for (i=0; user_modes[i].mode != '\0'; i++)
    {
      if (*pos == user_modes[i].mode)
      {
        matched = TRUE;
        if (what == MTYPE_ADD)
	{
          u->modes |= user_modes[i].value;
          /*log (1, "MODE %s :+%s", u->nick, user_modes[i].mode);*/
        }
        else
          u->modes &= ~user_modes[i].value;
        break;
      }
    }

    if (matched)
      continue;
  }
  return;
}

void SendQuit (char *reason)
{
	struct Bot *bot;
	dlink_node *dl;

#ifdef NICKSERVICES
    sts (":%s QUIT :%s", NickServ.nick, reason);

#ifdef CHANNELSERVICES
    sts (":%s QUIT :%s", ChanServ.nick, reason);
#endif

#ifdef MEMOSERVICES
    sts (":%s QUIT :%s", MemoServ.nick, reason);
#endif

#ifdef BOTSERVICES
    sts (":%s QUIT :%s", BotServ.nick, reason);

	DLINK_FOREACH (dl, BotServ.bots.head)
	{
		bot = (struct Bot *)dl->data;
		sts (":%s QUIT :%s", bot->nick, reason);
	}
#endif
#endif

    sts (":%s QUIT :%s", OperServ.nick, reason);
}
