#include "fwserv.h"
#include "config.h"
#include "nickserv.h"
#include "operserv.h"
#include "botserv.h"
#include "user.h"
#include "channel.h"

#include <mysql.h>

char *ChanModes (struct NewChannel *);

void utilserv(struct User *, char *, char *param);

static int u_scan(struct User *, int, char **);
static int u_quit(struct User *, int, char **);
static int u_login(struct User *, int, char **);
static int u_help(struct User *, int, char **);
static int u_chans(struct User *, int, char **);
static int u_nicks(struct User *, int, char **);
static int u_services (struct User *, int, char **);
static int u_regnicks (struct User *, int, char **);
static int u_settings (struct User *, int, char **);
static int u_opers (struct User *, int, char **);
static int u_events (struct User *, int, char **);
static int u_stats (struct User *, int, char **);
static int u_mem (struct User *, int, char **);
static int u_bots (struct User *, int, char **);
static int u_scanned (struct User *, int, char **);
static int u_servers (struct User *, int, char **);

static struct CmdTable util_table[] =
{
  { "SCAN",	NULL,	u_scan,		LVL_ADMIN },
  { "QUIT",	NULL,	u_quit,		LVL_ADMIN },
  { "LOGIN",	NULL,	u_login,	LVL_ADMIN },
  { "HELP",	NULL,	u_help,		LVL_ADMIN },
  { "CHANNELS",	NULL,	u_chans,	LVL_ADMIN },
  { "NICKS",	NULL,	u_nicks,	LVL_ADMIN },
  { "SERVICES",	NULL,	u_services,	LVL_ADMIN },
  { "REGNICKS",	NULL,	u_regnicks,	LVL_ADMIN },
  { "SETTINGS",	NULL,	u_settings,	LVL_ADMIN },
  { "OPERS",	NULL,	u_opers,	LVL_ADMIN },
  { "EVENTS",	NULL,	u_events,	LVL_ADMIN },
  { "STATS",	NULL,	u_stats,	LVL_ADMIN },
  { "MEM",	NULL,	u_mem,		LVL_ADMIN },
  { "BOTS",	NULL,	u_bots,		LVL_ADMIN },
  { "SCANNED",	NULL,	u_scanned,	LVL_ADMIN },
  { "SERVERS",	NULL,	u_servers,	LVL_ADMIN },
  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0 }
};

static int
u_servers (struct User *u, int parc, char *parv[])
{
  struct Server *server;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, server_list.head)
  {
    server = (struct Server *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: %s", count, server->name);
  }
}

static int
u_scanned (struct User *u, int parc, char *parv[])
{
  struct Scanned *scanned;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, scanned_list.head)
  {
    scanned = (struct Scanned *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: %s", count, scanned->ip);
  }
} 

static int
u_mem (struct User *u, int parc, char *parv[])
{
  notice(u, n_UtilServ, "dlink_node count: %d", Mem.dlink_count);
  notice(u, n_UtilServ, "nick count: %d", Mem.nick_count);
}

static int
u_stats (struct User *u, int parc, char *parv[])
{
  struct tm*  tmptr;
  static char timebuffer[MAX_DATE_STRING];

  tmptr = localtime((const time_t*)&me.started);
  strftime(timebuffer, MAX_DATE_STRING, "%Y/%m/%d %H:%M:%S", tmptr);
  notice(u, n_UtilServ, "Started: %s", timebuffer);
}

static int
u_events (struct User *u, int parc, char *parv[])
{
  show_events(u->nick);
}

static int
u_opers (struct User *u, int parc, char *parv[])
{
/*
  dlink_node *dl;
  struct MyOper *Oper;
  int count = 0;

  DLINK_FOREACH(dl, myopers_list.head)
  {
    Oper = (struct MyOper *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: Oper: %s", count, Oper->nick);
  }

  if (count == 0)
    notice(u, n_UtilServ, "No Opers");
*/
}

static int
u_bots (struct User *u, int parc, char *parv[])
{
  dlink_node *dl;
  struct Bots *bots;
  int count = 0;

  DLINK_FOREACH(dl, bot_list.head)
  {
    count++;
    bots = (struct Bots *)dl->data;
    notice(u, n_UtilServ, "%d: %s!%s@%s [%s] (%s)", count, bots->nick, bots->username, bots->hostname, bots->realhost, bots->gecos);
  }
}

static int
u_settings (struct User *u, int parc, char *parv[])
{
  notice(u, n_UtilServ, "Server Name: %s", me.name);
  notice(u, n_UtilServ, "Server Description: %s", me.desc);
  notice(u, n_UtilServ, "Uplink: %s", me.uplink);
  notice(u, n_UtilServ, "Uplink Pass: %s", me.pass);
  notice(u, n_UtilServ, "Uplink Port: %d", me.port);
  return;
}

static int
u_services (struct User *u, int parc, char *parv[])
{
  struct Services *service;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, services_list.head)
  {
    count++;
    service = (struct Services *)dl->data;
    notice(u, n_UtilServ, "%d: %s (%s)", count, service->nick, service->realname);
  }
  return;
}

static int
u_nicks (struct User *u, int parc, char *parv[])
{
  struct User *user;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, user_list.head)
  {
    user = (struct User *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: %s (%s) Connected via %s", count, user->nick, user->hostname, user->server->name);
  }
  return;
}

static int
u_chans (struct User *u, int parc, char *parv[])
{
  struct Channel *chan;
  struct NewChannel *c;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, channel_list.head)
  {
    c = (struct NewChannel *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: %s %s [%d] (%s)", count, c->name, (c->founder)? c->founder->nick : "", c->ts, ChanModes(c));
  }
  return;
}

static int
u_regnicks (struct User *u, int parc, char *parv[])
{
  struct Nick *nick;
  dlink_node *dl;
  int count = 0;

  DLINK_FOREACH(dl, nick_list.head)
  {
    nick = (struct Nick *)dl->data;
    count++;
    notice(u, n_UtilServ, "%d: %s (%s) [%s]", count, nick->nick, nick->password, nick->last_quit);
  }

  if (count == 0)
    notice(u, n_UtilServ, "No Registered Nicks"); 
  return;
}

static int
u_help (struct User *u, int parc, char *parv[])
{
  notice(u, n_UtilServ, "No Help Available");
  return;
}

static int
u_login(struct User *u, int parc, char *parv[])
{
  if (parc != 3)
  {
    notice(u, n_UtilServ, "Syntax: LOGIN <nick> <password>");
    return;
  }

  if (!strcasecmp(parv[1], "jeremy"))
  {
    if (!strcasecmp(parv[2], "sept1984"))
    {
      u->flags |= FLAG_AUTH;
      notice(u, n_UtilServ, "%s Logged in", u->nick);
    }
  }
  return;
}

static int
u_scan(struct User *u, int parc, char *parv[])
{
  if (parc != 2)
  {
    notice(u, n_UtilServ, "Syntax: SCAN <ip>");
    return;
  }

#ifdef PROXYSCAN
  ProxyScan(parv[1]);
#endif
  return;
}

static int
u_quit(struct User *u, int parc, char *parv[])
{
  char reason[BUFSIZE];

  if (!IsAuth(u))
  {
    notice(u, n_UtilServ, "Access Denied");
    return;
  }

  if (parc != 2)
    sprintf(reason, "Shutdown: Request by %s", u->nick);
  else
    sprintf(reason, "Shutdown: %s", parv[1]);

  sts(":%s QUIT :%s", "UtilServ", reason);
  sts(":%s QUIT :%s", "MChanServ", reason);
  close(uplinksock);
  del_connection(uplinksock);
  uplinksock = -1;
  log(1, "Shutdown requested by: %s", u->nick);
  sleep(1);
  exit(0);
}

void utilserv(struct User *u, char *target, char *param)
{
  char *parv[MAXPARA];
  int parc, i, status;
  static char buf[BUFSIZE];
  parc = 0;

  for (i = 0; i < MAXPARA; i++)
    parv[i] = NULL;

  strlcpy(buf, param, BUFSIZE);

  if (param != NULL)
    parc = msgtoken(param, parv);

  if (parc == 0)
    return;

  for (i = 0; util_table[i].handler; i++)
  {
    if(!strcasecmp(parv[0], util_table[i].name))
    {
      status = util_table[i].handler(u, parc, parv);
      return;
    }
  }

  notice(u, n_UtilServ, "Invalid Command.");
  return;
}

char *ChanModes (struct NewChannel *c)
{
  static char string[BUFSIZE];
  string[0] = '\0';

  if (c->modes & CMODE_INVITE)
    if (string[0] == '\0')
      strcpy(string, "i");
    else
      strcat(string, "i");

  if (c->modes & CMODE_MOD)
    if (string[0] == '\0')
      strcpy(string, "m");
    else
      strcat(string, "m");

  if (c->modes & CMODE_NOEXT)
    if (string[0] == '\0')
      strcpy(string, "n");
    else
      strcat(string, "n");

  if (c->modes & CMODE_TOPIC)
    if (string[0] == '\0')
      strcpy(string, "t");
    else
      strcat(string, "t");

  if (c->modes & CMODE_SEC)
    if (string[0] == '\0')
      strcpy(string, "s");
    else
      strcat(string, "s");

  if (c->modes & CMODE_PRIV)
    if (string[0] == '\0')
      strcpy(string, "p");
    else
      strcat(string, "p");

  if (c->modes & CMODE_OPERONLY)
    if (string[0] == '\0')
      strcpy(string, "O");
    else
      strcat(string, "O");

  return string;
}
